from preql.core.pql_types import T

DOCS = {
T.any: """A meta-type that can match any type.

	Examples:
		>> isa(my_obj, any)		// always returns true
		true
		>> isa(my_type, any)	// always returns true
		true
""",

# T.unknown: None,	# TODO

T.union: """A meta-type that means 'either one of the given types'

	Example:
		>> int <= union[int, string]
		true
		>> union[int, string] <= int
		false
		>> union[int, string] <= union[string, int]
		true

""",

T.type: """The type of types

	Examples:
		>> type(int) == type(string)
		true
		>> int <= type(int)			# int isn't a subtype of `type`
		false
		>> isa(int, type(int))		# int is an instance of `type`
		true
""",
T.object: """The base object type
""",

T.nulltype: """The type of the singleton `null`.

	Represents SQL `NULL`, but behaves like Python's `None`, 

	Examples:
		>> null == null
		true

		>> null + 1
		TypeError: Operator '+' not implemented for nulltype and int
""",

T.primitive: "The base type for all primitives",

T.text: "A text type (behaves the same as `string`)",
T.string: "A string type (behaves the same as `text`)",
T.number: "The base type for all numbers",
T.int: "An integer number",
T.float: "A floating-point number",
T.bool: "A boolean, which can be either `true` or `false`",
# T.decimal: "A decimal number",

T.datetime: "A datetime type",

T.container: """The base type of containers.

	A container holds other objects inside it. 
""",

T.struct: "A structure type",

T.row: "A row in a table. (essentially a named-tuple)",

# T.collection: """The base class of collections.

# 	A collection holds an array of other objects inside it.
# """,

T.table: """A table type.

	Tables support the following operations -
	- Projection (or: map), using the `{}` operator
	- Selection (or: filter), using the `[]` operator
	- Slice (or: indexing), using the `[..]` operator
	- Order (or: sorting), using the `order{}` operator
	- Update, using the `update{}` operator
	- Delete, using the `delete[]` operator
	- `+` for concat, `&` for intersect, `|` for union
""",

T.list: """A list type""",
T.set: """A set type, in which all elements are unique""",
T.projected: """A meta-type to signify projected operations, i.e. operations inside a projection.

	Example:
		>> x = [1]
		>> one one x{ repr(type(item)) }
		"projected[item: int]"
""",
T.aggregated: """A meta-type to signify aggregated operations, i.e. operations inside a grouping

	Example:
		>> x = [1]
		>> one one x{ => repr(type(item))}
		"aggregated[item: int]"
""",
T.t_id: "The type of a table id",
T.t_relation: "The type of a table relation",


T.json: "A json type",
T.json_array: "A json array type. Created by aggregation.",

T.function: "A meta-type for all functions",
T.module: "A meta-type for all modules",
T.signal: "A meta-type for all signals (i.e. exceptions)",

}




import sys

optimize = True
cache = False
debug = False

print_sql = False
typecheck = False # not sys.flags.optimize
autocomplete = True

try:
    from .local_settings import *
except ImportError:
    pass


from contextlib import contextmanager
from copy import copy
from logging import getLogger

from preql.utils import dsp
from preql.sql_interface import ConnectError, create_engine

from . import pql_ast as ast
from . import pql_objects as objects
from .exceptions import Signal
from .exceptions import InsufficientAccessLevel
from .pql_types import Type, T

logger = getLogger('interp')

# Define common dispatch functions
@dsp
def simplify(state, obj: type(NotImplemented)) -> object:
    raise NotImplementedError()

@dsp
def evaluate(state, obj: type(NotImplemented)) -> object:
    raise NotImplementedError()

@dsp
def cast_to_python(state, obj: type(NotImplemented)) -> object:
    raise NotImplementedError(obj)



class AccessLevels:
    COMPILE = 1
    # COMPILE_TEXT
    QUERY = 2
    EVALUATE = 3
    READ_DB = 4
    WRITE_DB = 5


class State:
    AccessLevels = AccessLevels

    def __init__(self, interp, db, display, ns=None):
        self.db = db
        self.interp = interp
        self.display = display
        # Add logger?

        self.ns = Namespace(ns)
        self.tick = [0]

        self.access_level = AccessLevels.WRITE_DB
        self._cache = {}
        self.stacktrace = []

    @classmethod
    def clone(cls, inst):
        s = cls(inst.interp, inst.db, inst.display)
        s.ns = copy(inst.ns)
        s.tick = inst.tick
        s.access_level = inst.access_level
        s._cache = inst._cache
        s.stacktrace = copy(inst.stacktrace)
        return s

    def __copy__(self):
        return self.clone(self)

    def limit_access(self, new_level):
        return self.reduce_access(min(new_level, self.access_level))

    def reduce_access(self, new_level):
        assert new_level <= self.access_level
        s = copy(self)
        s.access_level = new_level
        return s

    def require_access(self, level):
        if self.access_level < level:
            raise InsufficientAccessLevel(level)
    def catch_access(self, level):
        if self.access_level < level:
            raise Exception("Bad access. Security risk.")

    def connect(self, uri, auto_create=False):
        logger.info(f"[Preql] Connecting to {uri}")
        try:
            self.db = create_engine(uri, self.db._print_sql, auto_create)
        except NotImplementedError as e:
            raise Signal.make(T.NotImplementedError, None, *e.args) from e
        except ConnectError as e:
            raise Signal.make(T.DbConnectionError, None, *e.args) from e
        except ValueError as e:
            raise Signal.make(T.ValueError, None, *e.args) from e

        self._db_uri = uri


    def get_all_vars(self):
        return self.ns.get_all_vars()

    def get_all_vars_with_rank(self):
        return self.ns.get_all_vars_with_rank()

    def get_var(self, name):
        try:
            return self.ns.get_var(name)
        except NameNotFound:
            builtins = self.ns.get_var('__builtins__')
            assert isinstance(builtins, objects.Module)
            try:
                return builtins.namespace[name]
            except KeyError:
                pass

            raise Signal.make(T.NameError, name, f"Name '{name}' is not defined")


    def set_var(self, name, value):
        return self.ns.set_var(name, value)

    def use_scope(self, scope: dict):
        return self.ns.use_scope(scope)

    def unique_name(self, obj):
        self.tick[0] += 1
        return obj + str(self.tick[0])


class NameNotFound(Exception):
    pass

class Namespace:
    def __init__(self, ns=None):
        self._ns = ns or [{}]

    def __copy__(self):
        return Namespace([dict(n) for n in self._ns])

    def get_var(self, name):
        for scope in reversed(self._ns):
            if name in scope:
                return scope[name]

        raise NameNotFound(name)

    def set_var(self, name, value):
        assert not isinstance(value, ast.Name)
        self._ns[-1][name] = value


    @contextmanager
    def use_scope(self, scope: dict):
        x = len(self._ns)
        self._ns.append(scope)
        try:
            yield
        finally:
            _discarded_scope = self._ns.pop()
            assert x == len(self._ns)


    # def push_scope(self):
    #     self.ns.append({})

    # def pop_scope(self):
    #     return self.ns.pop()

    def __len__(self):
        return len(self._ns)

    def get_all_vars(self):
        d = {}
        for scope in reversed(self._ns):
            d.update(scope) # Overwrite upper scopes
        return d

    def get_all_vars_with_rank(self):
        d = {}
        for i, scope in enumerate(reversed(self._ns)):
            for k, v in scope.items():
                if k not in d:
                    d[k] = i, v
        return d



def assert_type(t, type_, ast_node, op, msg="%s expected an object of type %s, instead got '%s'"):
    assert isinstance(t, Type), t
    assert isinstance(type_, Type)
    if not t <= type_:
        if type_.typename == 'union':
            type_str = ' or '.join("'%s'" % elem for elem in type_.elems)
        else:
            type_str = "'%s'" % type_
        raise Signal.make(T.TypeError, ast_node, msg % (op, type_str, t))

def exclude_fields(state, table, fields):
    proj = ast.Projection(table, [ast.NamedField(None, ast.Ellipsis(None, exclude=list(fields) ), user_defined=False)])
    return evaluate(state, proj)

def call_builtin_func(state, name, args):
    "Call a builtin pql function"
    builtins = state.ns.get_var('__builtins__')
    assert isinstance(builtins, objects.Module)

    expr = ast.FuncCall(builtins.namespace[name], args)
    return evaluate(state, expr)



def is_global_scope(state):
    assert len(state.ns) != 0
    return len(state.ns) == 1


# def cast_to_python_primitive(state, obj):
#     res = cast_to_python(state, obj)
#     assert isinstance(res, (int, str, float, dict, list, type(None), datetime)), (res, type(res))
#     return res

def cast_to_python_string(state, obj: objects.AbsInstance):
    res = cast_to_python(state, obj)
    if not isinstance(res, str):
        raise Signal.make(T.TypeError, obj, f"Expected string, got '{res}'")
    return res

def cast_to_python_int(state, obj: objects.AbsInstance):
    res = cast_to_python(state, obj)
    if not isinstance(res, int):
        raise Signal.make(T.TypeError, obj, f"Expected string, got '{res}'")
    return res

    
pyvalue_inst = objects.pyvalue_inst




from pathlib import Path
import subprocess
import json

import dsnparse

from .utils import classify, dataclass
from .loggers import sql_log
from .context import context

from .core.sql import Sql, QueryBuilder, sqlite, postgres, mysql, duck, from_sql, bigquery, _quote
from .core.pql_types import T, Type, Object, Id
from .core.exceptions import DatabaseQueryError, Signal

@dataclass
class Const(Object):
    type: Type
    value: object

class ConnectError(Exception):
    pass


def log_sql(sql):
    for i, s in enumerate(sql.split('\n')):
        prefix = '/**/    ' if i else '/**/;;  '
        sql_log.debug(prefix+s)


class SqlInterface:
    _conn: object

    def __init__(self, print_sql=False):
        self._print_sql = print_sql

    def query(self, sql, subqueries=None, qargs=(), quiet=False):
        assert context.state
        assert isinstance(sql, Sql), sql
        sql_code = self.compile_sql(sql, subqueries, qargs)

        if self._print_sql and not quiet:
            log_sql(sql_code)

        return self._execute_sql(sql.type, sql_code)
        # return self._import_result(sql.type, cur, state)

    def _import_result(self, sql_type, c):
        if sql_type is not T.nulltype:
            res = c.fetchall()
            return from_sql(Const(sql_type, res))



    def compile_sql(self, sql, subqueries=None, qargs=()):
        qb = QueryBuilder(self.target)

        return sql.finalize_with_subqueries(qb, subqueries)

    def commit(self):
        self._conn.commit()

    def rollback(self):
        self._conn.rollback()

    def close(self):
        self._conn.close()

    def import_table_types(self):
        tables = self.list_tables()
        for table_name in tables:
            table_type = self.import_table_type(table_name)
            yield None, table_name, table_type

    def qualified_name(self, name):
        return name


class SqlInterfaceCursor(SqlInterface):
    "An interface that uses the standard SQL cursor interface"

    def _backend_execute_sql(self, sql_code):
        c = self._conn.cursor()
        c.execute(sql_code)
        return c

    def _execute_sql(self, sql_type, sql_code):
        assert context.state
        try:
            c = self._backend_execute_sql(sql_code)
        except Exception as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise DatabaseQueryError(msg%(sql_code, e))

        return self._import_result(sql_type, c)

    def ping(self):
        c = self._conn.cursor()
        c.execute('select 1')
        row ,= c.fetchall()
        n ,= row
        assert n == 1


class MysqlInterface(SqlInterfaceCursor):
    target = mysql

    def __init__(self, host, port, database, user, password, print_sql=False):
        import mysql.connector
        from mysql.connector import errorcode

        args = dict(host=host, port=port, database=database, user=user, password=password)
        args = {k:v for k, v in args.items() if v is not None}

        try:
            # TODO utf8??
            self._conn = mysql.connector.connect(charset='utf8', use_unicode=True, **args)
        except mysql.connector.Error as e:
            if e.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                raise ConnectError("Bad user name or password") from e
            elif e.errno == errorcode.ER_BAD_DB_ERROR:
                raise ConnectError("Database does not exist") from e
            else:
                raise ConnectError(*e.args) from e

        self._print_sql = print_sql

    def table_exists(self, name):
        tables = [t.lower() for t in self.list_tables()]
        return name.lower() in tables

    def list_tables(self):
        sql_code = "SHOW TABLES"
        return self._execute_sql(T.list[T.string], sql_code)

    def import_table_type(self, name, columns_whitelist=None):
        columns_t = T.table(dict(
            name=T.string,
            type=T.string,
            nullable=T.string,
            key=T.string,
            default=T.string,
            extra=T.string,
        ))
        columns_q = "desc %s" % name
        sql_columns = self._execute_sql(columns_t, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = {c['name']: _type_from_sql(c['type'].decode(), c['nullable']) for c in sql_columns}

        return T.table(cols, name=Id(name))


class PostgresInterface(SqlInterfaceCursor):
    target = postgres

    def __init__(self, host, port, database, user, password, print_sql=False):
        import psycopg2
        import psycopg2.extras
        psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)
        try:
            self._conn = psycopg2.connect(host=host, port=port, database=database, user=user, password=password)
        except psycopg2.OperationalError as e:
            raise ConnectError(*e.args) from e

        self._print_sql = print_sql


    def table_exists(self, name):
        sql_code = "SELECT count(*) FROM information_schema.tables where table_name='%s'" % name
        cnt = self._execute_sql(T.int, sql_code)
        return cnt > 0

    def list_tables(self):
        sql_code = "SELECT table_name FROM information_schema.tables where table_schema='public'"
        return self._execute_sql(T.list[T.string], sql_code)


    _schema_columns_t = T.table(dict(
        schema=T.string,
        table=T.string,
        name=T.string,
        pos=T.int,
        nullable=T.bool,
        type=T.string,
    ))

    def import_table_type(self, name, columns_whitelist=None):

        columns_q = """SELECT table_schema, table_name, column_name, ordinal_position, is_nullable, data_type
            FROM information_schema.columns
            WHERE table_name = '%s'
            """ % name
        sql_columns = self._execute_sql(self._schema_columns_t, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['pos'], c['name'], _type_from_sql(c['type'], c['nullable'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        return T.table(cols, name=Id(name))

    def import_table_types(self):
        columns_q = """SELECT table_schema, table_name, column_name, ordinal_position, is_nullable, data_type
            FROM information_schema.columns
            """
        sql_columns = self._execute_sql(self._schema_columns_t, columns_q)

        columns_by_table = classify(sql_columns, lambda c: (c['schema'], c['table']))

        for (schema, table_name), columns in columns_by_table.items():
            cols = [(c['pos'], c['name'], _type_from_sql(c['type'], c['nullable'])) for c in columns]
            cols.sort()
            cols = dict(c[1:] for c in cols)

            # name = '%s.%s' % (schema, table_name)
            yield schema, table_name, T.table(cols, name=Id(schema, table_name))



class BigQueryInterface(SqlInterface):
    target = bigquery

    def __init__(self, project, print_sql=False):
        from google.cloud import bigquery

        self._client = bigquery.Client(project)
        self._default_dataset = None

        self._print_sql = print_sql


    def table_exists(self, name):
        from google.api_core.exceptions import NotFound
        try:
            self._client.get_table(name)
        except NotFound:
            return False
        return True

    def _list_tables(self):
        for ds in self._client.list_datasets():
            for t in self._client.list_tables(ds.reference):
                yield t.full_table_id.replace(':', '.')     # Hacky

    def list_tables(self):
        return list(self._list_tables())


    def _execute_sql(self, sql_type, sql_code):
        assert context.state
        try:
            res = list(self._client.query(sql_code))
        except Exception as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise DatabaseQueryError(msg%(sql_code, e))

        if sql_type is not T.nulltype:
            res = [list(i.values()) for i in res]
            return from_sql(Const(sql_type, res))



    _schema_columns_t = T.table(dict(
        schema=T.string,
        table=T.string,
        name=T.string,
        pos=T.int,
        nullable=T.bool,
        type=T.string,
    ))

    def import_table_type(self, name, columns_whitelist=None):
        cols = {}
        for f in self._client.get_table(name).schema:
            if columns_whitelist is None or f.name in columns_whitelist:
                cols[f.name] = _type_from_sql(f.field_type, f.is_nullable)

        return T.table(cols, name=Id(name))

    def list_datasets(self):
        return [ds.dataset_id for ds in self._client.list_datasets()]

    def set_default_dataset(self, dataset):
        self._default_dataset = dataset

    def get_default_dataset(self):
        if self._default_dataset is None:
            datasets = self.list_datasets()
            if not datasets:
                raise Signal(T.ValueError, None, "No dataset found.")
            self._default_dataset = datasets[0]
        return self._default_dataset

    def qualified_name(self, name):
        if '.' in name: # already has dataset
            return name
        return self.get_default_dataset() + '.' + name

    def rollback(self):
        # XXX No error? No warning?
        pass
    def commit(self):
        # XXX No error? No warning?
        pass

class AbsSqliteInterface:
    def table_exists(self, name):
        sql_code = "SELECT count(*) FROM sqlite_master where name='%s' and type='table'" % name
        cnt = self._execute_sql(T.int, sql_code)
        return cnt > 0

    def list_tables(self):
        sql_code = "SELECT name FROM sqlite_master where type='table'"
        return self._execute_sql(T.list[T.string], sql_code)

    table_schema_type = T.table(dict(
        pos=T.int,
        name=T.string,
        type=T.string,
        notnull=T.bool,
        default_value=T.string,
        pk=T.bool,
    ))

    def import_table_type(self, name, columns_whitelist=None):
        columns_q = """pragma table_info('%s')""" % name
        sql_columns = self._execute_sql(self.table_schema_type, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['pos'], c['name'], _type_from_sql(c['type'], not c['notnull'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        pk = [[c['name']] for c in sql_columns if c['pk']]

        return T.table(cols, name=Id(name), pk=pk)


class SqliteInterface(SqlInterfaceCursor, AbsSqliteInterface):
    target = sqlite

    def __init__(self, filename=None, print_sql=False):
        import sqlite3
        try:
            self._conn = sqlite3.connect(filename or ':memory:')
        except sqlite3.OperationalError as e:
            raise ConnectError(*e.args) from e

        def sqlite_power(x,n):
            return x**n
        self._conn.create_function("power", 2, sqlite_power)

        self._print_sql = print_sql



class DuckInterface(AbsSqliteInterface):
    target = duck

    def __init__(self, filename=None, print_sql=False):
        import duckdb
        self._conn = duckdb.connect(filename or ':memory:')
        self._print_sql = print_sql

    def rollback(self):
        pass    # XXX


class GitInterface(AbsSqliteInterface):
    "Uses https://github.com/augmentable-dev/askgit"

    target = sqlite

    def __init__(self, path, print_sql):
        self.path = path
        self._print_sql = print_sql


    table_schema_type = T.table(dict(
        cid=T.int,
        dflt_value=T.string,
        name=T.string,
        notnull=T.bool,
        pk=T.bool,
        type=T.string,
    ))

    def _execute_sql(self, sql_type, sql_code):
        assert context.state
        try:
            res = subprocess.check_output(['askgit', '--format', 'json', sql_code])
        except FileNotFoundError:
            msg = "Could not find executable 'askgit'. Make sure it's installed, and try again."
            raise DatabaseQueryError(msg)
        except subprocess.CalledProcessError as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise DatabaseQueryError(msg%(sql_code, e))

        return self._import_result(sql_type, res)

    def _import_result(self, sql_type, c):
        if sql_type is not T.nulltype:
            if sql_type <= T.table:
                lookup = dict(reversed(x) for x in enumerate(sql_type.elems))
                rows = [json.loads(x) for x in c.split(b'\n') if x.strip()]
                res = []
                for row in rows:
                    # TODO refactor into a function
                    x = ["PLACEHOLDER"] * len(row)
                    for k, v in row.items():
                        x[lookup[k]] = v
                    assert "PLACEHOLDER" not in x, (x, row)
                    res.append(x)
            else:
                res = [list(json.loads(x).values()) for x in c.split(b'\n') if x.strip()]

            return from_sql(Const(sql_type, res))

    def import_table_type(self, name, columns_whitelist=None):
        # TODO merge with superclass

        columns_q = """pragma table_info('%s')""" % name
        sql_columns = self._execute_sql(self.table_schema_type, columns_q)

        if columns_whitelist:
            wl = set(columns_whitelist)
            sql_columns = [c for c in sql_columns if c['name'] in wl]

        cols = [(c['cid'], c['name'], _type_from_sql(c['type'], not c['notnull'])) for c in sql_columns]
        cols.sort()
        cols = dict(c[1:] for c in cols)

        return T.table(cols, name=Id(name))

    def list_tables(self):
        # TODO merge with superclass?
        sql_code = "SELECT name FROM sqlite_master where type='table'"
        res = self._execute_sql(T.table(dict(name=T.string)), sql_code)
        return [x['name'] for x in res]



def _bool_from_sql(n):
    if n == 'NO':
        n = False
    if n == 'YES':
        n = True
    assert isinstance(n, bool), n
    return n

def _type_from_sql(type, nullable):
    type = type.lower()
    d = {
        'integer': T.int,
        'int': T.int,           # mysql
        'tinyint(1)': T.bool,   # mysql
        'serial': T.t_id,
        'bigserial': T.t_id,
        'smallint': T.int,  # TODO smallint / bigint?
        'bigint': T.int,
        'character varying': T.string,
        'character': T.string,  # TODO char?
        'real': T.float,
        'float': T.float,
        'double precision': T.float,    # double on 32-bit?
        'boolean': T.bool,
        'timestamp': T.datetime,
        'timestamp without time zone': T.datetime,
        'text': T.text,
    }
    try:
        v = d[type]
    except KeyError:
        if type.startswith('int('): # TODO actually parse it
            return T.int
        elif type.startswith('tinyint('): # TODO actually parse it
            return T.int
        elif type.startswith('varchar('): # TODO actually parse it
            return T.string

        return T.string.as_nullable()

    nullable = _bool_from_sql(nullable)

    return v.replace(_nullable=nullable)


def _drop_tables(state, *tables):
    # XXX temporary. Used for testing
    db = state.db
    with context(state=state):
        for t in tables:
            t = _quote(db.target, db.qualified_name(t))
            db._execute_sql(T.nulltype, f"DROP TABLE {t};")


def create_engine(db_uri, print_sql, auto_create):
    dsn = dsnparse.parse(db_uri)
    if len(dsn.paths) != 1:
        raise ValueError("Bad value for uri: %s" % db_uri)
    path ,= dsn.paths
    if len(dsn.schemes) > 1:
        raise NotImplementedError("Preql doesn't support multiple schemes")
    scheme ,= dsn.schemes
    if scheme == 'sqlite':
        if not auto_create and path != ':memory:':
            if not Path(path).exists():
                raise ConnectError("File %r doesn't exist. To create it, set auto_create to True" % path)
        return SqliteInterface(path, print_sql=print_sql)
    elif scheme == 'postgres':
        return PostgresInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
    elif scheme == 'mysql':
        return MysqlInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
    elif scheme == 'git':
        return GitInterface(path, print_sql=print_sql)
    elif scheme == 'duck':
        return DuckInterface(path, print_sql=print_sql)
    elif scheme == 'bigquery':
        return BigQueryInterface(path, print_sql=print_sql)

    raise NotImplementedError(f"Scheme {dsn.scheme} currently not supported")



from runtype import dataclass

from preql.utils import safezip, dsp

from preql.docstring.docstring import parse, Section, Defin, Text

from preql.core.pql_objects import Module, Function, T
from preql.core.pql_types import Type, subtypes

from . import type_docs


class AutoDocError(Exception):
    pass

@dataclass
class ModuleDoc:
    module: object
    items: list

    def print_text(self):
        s = f'\n[dodger_blue2]module[/dodger_blue2] [bold white]{self.module.name}[/bold white]\n'
        line = '=' * (len(self.module.name) + 7)
        s += f'[dodger_blue2]{line}[/dodger_blue2]\n\n\n'
        return s + '\n\n'.join(i.print_text(2) for i in self.items)

    def print_rst(self):
        s = f'\n{self.module.name}\n'
        line = '-' * len(self.module.name)
        s += f'{line}\n\n\n'
        items = [i.print_rst() for i in self.items]
        return s + '\n\n'.join(items)


@dataclass
class FuncDoc:
    func: object
    doc: object

    def print_text(self, indent=0):
        params = [str(p.name) for p in self.func.params]
        if self.func.param_collector:
            params.append('...' + self.func.param_collector.name)
        params = ', '.join(params)
        indent_str = ' ' * indent
        s = f'{indent_str}[dodger_blue2]func[/dodger_blue2] [bold white]{self.func.name}[/bold white]({params}) = ...\n\n'
        return s + self.doc.print_text(indent+4)

    def print_rst(self):
        params = [str(p.name) for p in self.func.params]
        if self.func.param_collector:
            params.append('...' + self.func.param_collector.name)
        params = ', '.join(params)
        s = f".. function:: {self.func.name}({params})\n\n"
        return s + self.doc.print_rst()

@dataclass
class TypeDoc:
    type: object
    doc: object

    def print_text(self, indent=0):
        params = [str(p) for p in self.type.elems]
        params = ', '.join(params)
        if params:
            params = f'\\[{params}]'
        indent_str = ' ' * indent
        s = f'{indent_str}[dodger_blue2]type[/dodger_blue2] [bold white]{self.type.typename}[/bold white]{params}\n\n'
        return s + self.doc.print_text(indent+4)

    def print_rst(self):
        s = f".. class:: {str(self.type)}⁣\n\n"     # includes an invisible unicode separator to trick sphinx
        return s + self.doc.print_rst()



from lark import LarkError

def doc_func(f):
    try:
        doc_tree = parse(f.docstring or '')
    except LarkError as e:
        raise AutoDocError(f"Error in docstring of function {f.name}")


    assert {s.name for s in doc_tree.sections} <= {'Parameters', 'Example', 'Examples', 'Note', 'Returns', 'See Also'}, [s.name for s in doc_tree.sections]
    try:
        params_doc = doc_tree.get_section('Parameters')
    except KeyError:
        if f.params:
            params_doc = Section('Parameters', [Defin(p.name, None, str(p.type) if p.type else '') for p in f.params])
            doc_tree.sections.insert(0, params_doc)
    else:
        params = list(f.params)
        if f.param_collector:
            params.append(f.param_collector)
        if len(params) != len(params_doc.items):
            raise AutoDocError(f"Parameters don't match docstring in function {f}")

        for d, p in safezip(params_doc.items, params):
            assert d.name == p.name, (d.name, p.name)
            d.type = str(p.type) if p.type else ''
            d.default = p.default.repr() if p.default else ''

    return FuncDoc(f, doc_tree)


def doc_module(m):
    return ModuleDoc(m, list(map(doc_func, m.public_functions())))


# def test_func():
#     p = Preql()
#     rich.print(doc_func(p('bfs')).print_text())

# def test_module():
#     p = Preql()
#     rich.print(doc_module(p('__builtins__')).print_text())

def generate_rst(modules_fn, types_fn):
    from preql import Preql
    p = Preql()

    with open(types_fn, 'w', encoding='utf8') as f:
        print('Preql Types', file=f)
        print('===========', file=f)
        for t in T.values():
            try:
                print(autodoc(t).print_rst(), file=f)
            except NotImplementedError:
                pass

    with open(modules_fn, 'w', encoding='utf8') as f:
        print('Preql Modules', file=f)
        print('=============', file=f)
        print(doc_module(p('__builtins__')).print_rst(), file=f)
        p('import graph')
        print(doc_module(p('graph')).print_rst(), file=f)

@dsp
def autodoc(m: Module):
    return doc_module(m)

@dsp
def autodoc(f: Function):
    return doc_func(f)

@dsp
def autodoc(t: Type):
    try:
        docstr = type_docs.DOCS[t]
    except KeyError:
        raise NotImplementedError(t)
    try:
        doc_tree = parse(docstr)
    except LarkError as e:
        raise AutoDocError(f"Error in docstring of type {t}")

    assert {s.name for s in doc_tree.sections} <= {'Example', 'Examples', 'Note', 'See Also'}, [s.name for s in doc_tree.sections]

    if t in subtypes:
        subtypes_doc = Section('Subtypes', [Text([str(st) + ", "]) for st in subtypes[t]])
        doc_tree.sections.insert(0, subtypes_doc)


    if t.supertypes:
        supertypes_doc = Section('Supertypes', [Text([str(st)]) for st in t.supertypes])
        doc_tree.sections.insert(0, supertypes_doc)

    return TypeDoc(t, doc_tree)

# test_func()
# test_module()
if __name__ == '__main__':
    generate_rst('preql-modules.rst', 'preql-types.rst')



from . import pql_objects as objects
from . import sql
from .pql_types import T, dp_type, ITEM_NAME
from .types_impl import kernel_type
from .exceptions import Signal

@dp_type
def _cast(inst_type, target_type, inst):
    if inst_type <= target_type:
        return inst
    raise Signal.make(T.TypeError, None, f"Cast not implemented for {inst_type}->{target_type}")

@dp_type
def _cast(inst_type: T.list, target_type: T.list, inst):
    if inst is objects.EmptyList:
        return inst.replace(type=target_type)

    if inst_type.elem <= target_type.elem:
        return inst

    value = inst.get_column(ITEM_NAME)
    elem = _cast(value.type, target_type.elem, value)
    code = sql.Select(target_type, inst.code, [sql.ColumnAlias(elem.code, ITEM_NAME)])
    return inst.replace(code=code, type=T.list[elem.type])


@dp_type
def _cast(inst_type: T.aggregated, target_type: T.list, inst):
    res = _cast(inst_type.elem, target_type.elem, inst)
    return objects.aggregate(res)   # ??

@dp_type
def _cast(inst_type: T.table, target_type: T.list, inst):
    t = inst.type
    if len(t.elems) != 1:
        raise Signal.make(T.TypeError, None, f"Cannot cast {inst_type} to {target_type}. Too many columns")
    if not inst_type.elem <= target_type.elem:
        raise Signal.make(T.TypeError, None, f"Cannot cast {inst_type} to {target_type}. Elements not matching")

    (elem_name, elem_type) ,= inst_type.elems.items()
    code = sql.Select(T.list[elem_type], inst.code, [sql.ColumnAlias(sql.Name(elem_type, elem_name), ITEM_NAME)])

    return objects.TableInstance.make(code, T.list[elem_type], [inst])

@dp_type
def _cast(_inst_type: T.t_id, _target_type: T.int, inst):
    return inst.replace(type=T.int)

@dp_type
def _cast(_inst_type: T.int, target_type: T.t_id, inst):
    return inst.replace(type=target_type)

@dp_type
def _cast(_inst_type: T.union[T.float, T.bool], _target_type: T.int, inst):
    code = sql.Cast(T.int, inst.code)
    return objects.Instance.make(code, T.int, [inst])

@dp_type
def _cast(_inst_type: T.number, _target_type: T.bool, inst):
    code = sql.Compare('!=', [inst.code, sql.make_value(0)])
    return objects.Instance.make(code, T.bool, [inst])

@dp_type
def _cast(_inst_type: T.string, _target_type: T.bool, inst):
    code = sql.Compare('!=', [inst.code, sql.make_value('')])
    return objects.Instance.make(code, T.bool, [inst])

@dp_type
def _cast(_inst_type: T.string, _target_type: T.text, inst):
    return inst.replace(type=T.text)
@dp_type
def _cast(_inst_type: T.text, _target_type: T.string, inst):
    return inst.replace(type=T.string)
@dp_type
def _cast(_inst_type: T.string, _target_type: T.string, inst):     # Disambiguate text<->string due to inheritance
    return inst

@dp_type
def _cast(_inst_type: T.union[T.int, T.bool], _target_type: T.float, inst):
    code = sql.Cast(T.float, inst.code)
    return objects.Instance.make(code, T.float, [inst])

# @dp_type
# def _cast(state, inst_type: T.string, target_type: T.int, inst):
#     # TODO error on bad string?
#     code = sql.Cast(T.int, "int", inst.code)
#     return objects.Instance.make(code, T.int, [inst])

@dp_type
def _cast(_inst_type: T.primitive, _target_type: T.string, inst):
    code = sql.Cast(T.string, inst.code)
    return objects.Instance.make(code, T.string, [inst])

@dp_type
def _cast(_inst_type: T.t_relation, target_type: T.t_id, inst):
    # TODO verify same table? same type?
    return inst.replace(type=target_type)

@dp_type
def _cast(inst_type: T.t_relation, target_type: T.int, inst):
    if inst.type.elem <= T.int:
        return inst.replace(type=target_type)
    raise Signal.make(T.TypeError, None, f"Cast not implemented for {inst_type}->{target_type}")

def cast(obj, t):
    res = _cast(kernel_type(obj.type), t, obj)
    return objects.inherit_phantom_type(res, [obj])



import inspect
import re
import os
from datetime import datetime
import csv
import itertools

import rich.progress
import runtype

from preql.utils import safezip, listgen, re_split
from preql.docstring.autodoc import autodoc, AutoDocError

from .exceptions import Signal, ExitInterp
from . import pql_objects as objects
from . import pql_ast as ast
from . import sql
from .interp_common import State, pyvalue_inst, assert_type, cast_to_python_string, cast_to_python_int, cast_to_python
from .evaluate import evaluate, db_query, TableConstructor, new_table_from_expr, new_table_from_rows
from .pql_types import T, Type, Id
from .types_impl import join_names
from .casts import cast
from .compiler import cast_to_instance

def new_str(x):
    return pyvalue_inst(str(x), T.string)

def _pql_PY_callback(state: State, var: str):
    var = var.group()
    assert var[0] == '$'
    var_name = var[1:]
    obj = state.get_var(var_name)
    inst = evaluate(state, obj)

    if not isinstance(inst, objects.ValueInstance):
        raise Signal.make(T.TypeError, None, f"Cannot convert {inst} to a Python value")

    return str(inst.local_value)


def pql_PY(state: State, code_expr: T.string, code_setup: T.string.as_nullable() = objects.null):
    """Evaluate the given Python expression and convert the result to a Preql object

    Parameters:
        code_expr: The Python expression to evaluate
        code_setup: Setup code to prepare for the evaluation

    Note:
        This function is still experemental, and should be used with caution.

    Example:
        >> PY("sys.version", "import sys")
        "3.8.2 (tags/v3.8.2:7b3ab59, Feb 25 2020, 23:03:10)"
    """
    py_code = cast_to_python_string(state, code_expr)
    py_setup = cast_to_python_string(state, code_setup)

    py_code = re.sub(r"\$\w+", lambda m: _pql_PY_callback(state, m), py_code)

    if py_setup:
        try:
            res = exec(py_setup)
        except Exception as e:
            raise Signal.make(T.EvalError, code_expr, f"Python code provided returned an error: {e}")

    try:
        res = eval(py_code)
    except Exception as e:
        raise Signal.make(T.EvalError, code_expr, f"Python code provided returned an error: {e}")

    return objects.from_python(res)
    # return pyvalue_inst(res)

def pql_inspect_sql(state: State, obj: T.object):
    """Returns the SQL code that would be executed to evaluate the given object

    """
    if not isinstance(obj, objects.Instance):
        raise Signal.make(T.TypeError, None, f"inspect_sql() expects a concrete object. Instead got: {obj.type}")
    s = state.db.compile_sql(obj.code, obj.subqueries)
    return objects.ValueInstance.make(sql.make_value(s), T.text, [], s)


def pql_SQL(_state: State, result_type: T.union[T.table, T.type], sql_code: T.string):
    """Create an object with the given SQL evaluation code, and given result type.

    The object will only be evaluated when required by the program flow.

    Using $var_name in the code will embed it in the query. Both primitives and tables are supported.

    A special `$self` variable allows to perform recursion, if supported by the dialect.


    Parameters:
        result_type: The expected type of the result of the SQL query
        sql_code: The SQL code to be evaluated

    Example:
        >> ["a", "b"]{item: SQL(string, "$item || '!'")}
        table  =2
        ┏━━━━━━━┓
        ┃ item  ┃
        ┡━━━━━━━┩
        │ a!    │
        │ b!    │
        └───────┘

        >> x = ["a", "b", "c"]
        >> SQL(type(x), "SELECT item || '!' FROM $x")
        table  =3
        ┏━━━━━━━┓
        ┃ item  ┃
        ┡━━━━━━━┩
        │ a!    │
        │ b!    │
        │ c!    │
        └───────┘
    """
    # TODO optimize for when the string is known (prefetch the variables and return Sql)
    # .. why not just compile with parameters? the types are already known
    return ast.ParameterizedSqlCode(result_type, sql_code)

def pql_force_eval(state: State, expr: T.object):
    """Forces the evaluation of the given expression.

    Executes any db queries necessary.
    """
    return objects.pyvalue_inst( cast_to_python(state, expr) )

def pql_fmt(state: State, s: T.string):
    """Format the given string using interpolation on variables marked as `$var`

    Example:
        >> ["a", "b", "c"]{item: fmt("$item!")}
        table  =3
        ┏━━━━━━━┓
        ┃ item  ┃
        ┡━━━━━━━┩
        │ a!    │
        │ b!    │
        │ c!    │
        └───────┘
    """
    _s = cast_to_python_string(state, s)

    tokens = re_split(r"\$\w+", _s)
    string_parts = []
    for m, t in tokens:
        if m:
            assert t[0] == '$'
            obj = state.get_var(t[1:])
            inst = cast_to_instance(state, obj)
            as_str = cast(inst, T.string)
            string_parts.append(as_str)
        elif t:
            string_parts.append(objects.pyvalue_inst(t))

    if not string_parts:
        return objects.pyvalue_inst('')
    elif len(string_parts) == 1:
        return string_parts[0]

    a = string_parts[0]
    for b in string_parts[1:]:
        a = ast.BinOp("+", [a,b])

    return cast_to_instance(state, a)



def _canonize_default(d):
    return None if d is inspect._empty else d

def create_internal_func(fname, f):
    sig = inspect.signature(f)
    return objects.InternalFunction(fname, [
        objects.Param(pname, type_ if isinstance(type_, Type) else T.any,
                      _canonize_default(sig.parameters[pname].default))
        for pname, type_ in list(f.__annotations__.items())[1:]
    ], f)


def pql_brk_continue(state):
    "Continue the execution of the code (exit debug interpreter)"
    pql_exit(state, objects.null)


def pql_breakpoint(state: State):
    """Creates a Python breakpoint.

    For a Preql breakpoint, use debug() instead
    """
    breakpoint()
    return objects.null

def pql_debug(state: State):
    """Breaks the execution of the interpreter, and enters into a debug
    session using the REPL environment.

    Use `c()` to continue the execution.
    """
    py_api = state._py_api

    with state.use_scope(breakpoint_funcs):
        py_api.start_repl('debug> ')
    return objects.null



def pql_issubclass(state: State, a: T.type, b: T.type):
    """Checks if type 'a' is a subclass of type 'b'

    Examples:
        >> issubclass(int, number)
        true
        >> issubclass(int, table)
        false
        >> issubclass(list, table)
        true
    """
    assert_type(a.type, T.type, a, 'issubclass')
    assert_type(b.type, T.type, b, 'issubclass')
    assert isinstance(a, Type)
    assert isinstance(b, Type)
    return pyvalue_inst(a <= b, T.bool)

def pql_isa(state: State, obj: T.any, type: T.type):
    """Checks if the give object is an instance of the given type

    Examples:
        >> isa(1, int)
        true
        >> isa(1, string)
        false
        >> isa(1.2, number)
        true
        >> isa([1], table)
        true
    """
    assert_type(type.type, T.type, obj, 'isa')
    res = obj.isa(type)
    return pyvalue_inst(res, T.bool)

def _count(state, obj, table_func, name='count'):
    if obj is objects.null:
        code = sql.FieldFunc(name, sql.AllFields(T.any))
    elif obj.type <= T.table:
        code = table_func(obj.code)
    elif isinstance(obj, objects.StructInstance) and not obj.type <= T.aggregated:
        # XXX is count() even the right method for this?
        return pyvalue_inst(len(obj.attrs))

    elif obj.type <= T.projected[T.json_array]:
        code = sql.JsonLength(obj.code)
    else:
        if not obj.type <= T.aggregated:
            msg = f"Function '{name}' expected an aggregated list, but got '{obj.type}' instead. Did you forget to group?"
            raise Signal.make(T.TypeError, None, msg)

        obj = obj.primary_key()
        code = sql.FieldFunc('count', obj.code)

    return objects.Instance.make(code, T.int, [obj])

def pql_count(state: State, obj: T.container.as_nullable() = objects.null):
    """Count how many rows are in the given table, or in the projected column.

    If no argument is given, count all the rows in the current projection.

    Examples:
        >> count([0..10])
        10
        >> [0..10]{ => count() }
        table  =1
        ┏━━━━━━━┓
        ┃ count ┃
        ┡━━━━━━━┩
        │    10 │
        └───────┘
        >> [0..10]{ => count(item) }
        table  =1
        ┏━━━━━━━┓
        ┃ count ┃
        ┡━━━━━━━┩
        │    10 │
        └───────┘
    """
    return _count(state, obj, sql.CountTable)


def pql_temptable(state: State, expr: T.table, const: T.bool.as_nullable() = objects.null):
    """Generate a temporary table with the contents of the given table

    It will remain available until the database session ends, unless manually removed.

    Parameters:
        expr: the table expression to create the table from
        const: whether the resulting table may be changed or not.

    Note:
        A non-const table creates its own `id` field.
        Trying to copy an existing id field into it will cause a collision
    """
    # 'temptable' creates its own counting 'id' field. Copying existing 'id' fields will cause a collision
    # 'const table' doesn't
    const = cast_to_python(state, const)
    assert_type(expr.type, T.table, expr, 'temptable')

    name = state.unique_name("temp")    # TODO get name from table options

    return new_table_from_expr(state, name, expr, const, True)


def pql_get_db_type(state: State):
    """Returns a string representing the type of the active database.

    Example:
        >> get_db_type()
        "sqlite"
    """
    assert state.access_level >= state.AccessLevels.EVALUATE
    return pyvalue_inst(state.db.target, T.string)



def sql_bin_op(op, t1, t2, name):

    if not isinstance(t1, objects.CollectionInstance):
        raise Signal.make(T.TypeError, t1, f"First argument isn't a table, it's a {t1.type}")
    if not isinstance(t2, objects.CollectionInstance):
        raise Signal.make(T.TypeError, t2, f"Second argument isn't a table, it's a {t2.type}")

    # TODO Smarter matching?
    l1 = len(t1.type.elems)
    l2 = len(t2.type.elems)
    if l1 != l2:
        msg = f"Cannot {name} tables due to column mismatch (table1 has {l1} columns, table2 has {l2} columns)"
        raise Signal.make(T.TypeError, None, msg)

    for e1, e2 in zip(t1.type.elems.values(), t2.type.elems.values()):
        if not e2 <= e1:
            raise Signal.make(T.TypeError, None, f"Cannot {name}. Column types don't match: '{e1}' and '{e2}'")

    code = sql.TableArith(op, [t1.code, t2.code])

    return type(t1).make(code, t1.type, [t1, t2])

def pql_table_intersect(_state: State, t1: T.table, t2: T.table):
    "Intersect two tables. Used for `t1 & t2`"
    return sql_bin_op("INTERSECT", t1, t2, "intersect")

def pql_table_substract(state: State, t1: T.table, t2: T.table):
    "Substract two tables (except). Used for `t1 - t2`"
    if state.db.target is sql.mysql:
        raise Signal.make(T.NotImplementedError, t1, "MySQL doesn't support EXCEPT (yeah, really!)")
    return sql_bin_op("EXCEPT", t1, t2, "subtract")

def pql_table_union(_state: State, t1: T.table, t2: T.table):
    "Union two tables. Used for `t1 | t2`"
    return sql_bin_op("UNION", t1, t2, "union")

def pql_table_concat(_state: State, t1: T.table, t2: T.table):
    "Concatenate two tables (union all). Used for `t1 + t2`"
    if isinstance(t1, objects.EmptyListInstance):
        return t2
    if isinstance(t2, objects.EmptyListInstance):
        return t1
    return sql_bin_op("UNION ALL", t1, t2, "concatenate")


def _get_table(t):
    if isinstance(t, objects.SelectedColumnInstance):
        return t.parent

    if not isinstance(t, objects.CollectionInstance):
        raise Signal.make(T.TypeError, None, f"join() arguments must be tables")
    return t

def _join2(a, b):
    if isinstance(a, objects.SelectedColumnInstance) and isinstance(b, objects.SelectedColumnInstance):
        return [a, b]

    if not ((a.type <= T.table) and (b.type <= T.table)):
        a = a.type.repr()
        b = b.type.repr()
        raise Signal.make(T.TypeError, None, f"join() arguments must be of same type. Instead got:\n * {a}\n * {b}")

    return _auto_join(a, b)

def _auto_join(ta, tb):
    refs1 = _find_table_reference(ta, tb)
    refs2 = _find_table_reference(tb, ta)
    auto_join_count = len(refs1) + len(refs2)
    if auto_join_count < 1:
        raise NoAutoJoinFound(ta, tb)
    elif auto_join_count > 1:   # Ambiguity in auto join resolution
        raise Signal.make(T.JoinError, None, "Cannot auto-join: Several plausible relations found")

    if len(refs1) == 1:
        dst, src = refs1[0]
    elif len(refs2) == 1:
        src, dst = refs2[0]
    else:
        assert False

    return src, dst


def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=None):

    names = list(exprs_dict)
    exprs = [evaluate(state, value) for value in exprs_dict.values()]

    # Validation and edge cases
    for x in exprs:
        if not isinstance(x, objects.AbsInstance):
            raise Signal.make(T.TypeError, None, f"Unexpected object type: {x}")

    for e in exprs:
        if e is objects.EmptyList:
            raise Signal.make(T.TypeError, None, "Cannot join on an untyped empty list")

        if isinstance(e, objects.UnknownInstance):
            table_type = T.table({n: T.unknown for n in names})
            return objects.TableInstance.make(sql.unknown, table_type, [])

    # Initialization
    tables = [_get_table(x) for x in exprs]
    assert all((t.type <= T.table) for t in tables)

    structs = {name: T.struct(table.type.elems) for name, table in safezip(names, tables)}

    if nullable:
        # Update nullable for left/right/outer joins
        structs = {name: t.as_nullable() if n else t
                   for (name, t), n in safezip(structs.items(), nullable)}

    tables = [objects.alias_table_columns(t, n) for n, t in safezip(names, tables)]

    primary_keys = [[name] + pk
                    for name, t in safezip(names, tables)
                    for pk in t.type.options.get('pk', [])
                   ]
    table_type = T.table(structs, name=Id(state.unique_name("joinall" if joinall else "join")), pk=primary_keys)

    conds = []
    if joinall:
        for e in exprs:
            if not isinstance(e, objects.CollectionInstance):
                raise Signal.make(T.TypeError, None, f"joinall() expected tables. Got {e}")
    else:
        if len(exprs) < 2:
            raise Signal.make(T.TypeError, None, "join expected at least 2 arguments")

        joined_exprs = set()
        for (na, ta), (nb, tb) in itertools.combinations(safezip(names, exprs), 2):
            try:
                cols = _join2(ta, tb)
                cond = sql.Compare('=', [sql.Name(c.type, join_names((n, c.name))) for n, c in safezip([na, nb], cols)])
                conds.append(cond)
                joined_exprs |= {id(ta), id(tb)}
            except NoAutoJoinFound as e:
                pass

        if {id(e) for e in exprs} != set(joined_exprs):
            # TODO better error!!! table name?? specific failed auto-join?
            s = ', '.join(repr(t.type) for t in exprs)
            raise Signal.make(T.JoinError, None, f"Cannot auto-join: No plausible relations found between {s}")


    code = sql.Join(table_type, join, [t.code for t in tables], conds)
    return objects.TableInstance.make(code, table_type, exprs)


def pql_join(state, tables):
    """Inner-join any number of tables.

    Each argument is expected to be one of -
    (1) A column to join on. Columns are attached to specific tables. or
    (2) A table to join on. The column will be chosen automatically, if there is no ambiguity.
    Connections are made according to the relationships in the declaration of the table.

    Parameters:
        tables: Each argument must be either a column, or a table.

    Returns:
        A new table, where each column is a struct representing one of
        the joined tables.

    Examples:
        >> join(a: [0].item, b: [0].item)
        table join46 =1
        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┓
        ┃ a           ┃ b           ┃
        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━┩
        │ {'item': 0} │ {'item': 0} │
        └─────────────┴─────────────┘

        >> join(a: [1..5].item, b: [3..8].item) {...a}
        table  =2
        ┏━━━━━━━┓
        ┃  item ┃
        ┡━━━━━━━┩
        │     3 │
        │     4 │
        └───────┘

        >> join(c: Country, l: Language) {...c, language: l.name}
    """
    return _join(state, "JOIN", tables)
def pql_leftjoin(state, tables):
    """Left-join any number of tables

    See `join`
    """
    return _join(state, "LEFT JOIN", tables, nullable=[False, True])
def pql_outerjoin(state, tables):
    """Outer-join any number of tables

    See `join`
    """
    return _join(state, "FULL OUTER JOIN", tables, nullable=[False, True])
def pql_joinall(state: State, tables):
    """Cartesian product of any number of tables

    See `join`

    Example:
        >> joinall(a: [0,1], b: ["a", "b"])
            table joinall14 =4
        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
        ┃ a           ┃ b             ┃
        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
        │ {'item': 0} │ {'item': 'a'} │
        │ {'item': 0} │ {'item': 'b'} │
        │ {'item': 1} │ {'item': 'a'} │
        │ {'item': 1} │ {'item': 'b'} │
        └─────────────┴───────────────┘
    """
    return _join(state, "JOIN", tables, True)

class NoAutoJoinFound(Exception):
    pass


@listgen
def _find_table_reference(t1, t2):
    for name, c in t1.type.elems.items():
        if c <= T.t_relation:
            rel = c.options['rel']
            if rel['table'] == t2.type:     # if same table
                yield t2.get_attr(rel['column']), objects.SelectedColumnInstance(t1, c, name)

def pql_type(_state: State, obj: T.any):
    """Returns the type of the given object

    Example:
        >> type(1)
        int
        >> type([1])
        list[item: int]
        >> type(int)
        type
    """
    return obj.type


def pql_repr(state: State, obj: T.any):
    """Returns the representation text of the given object
    """
    if obj.type <= T.projected | T.aggregated:
        raise Signal.make(T.CompileError, obj, "repr() cannot run in projected/aggregated mode")

    try:
        return pyvalue_inst(obj.repr())
    except ValueError:
        value = repr(cast_to_python(state, obj))
        return pyvalue_inst(value)

def pql_columns(_state: State, obj: T.container):
    """Returns a dictionary `{column_name: column_type}` for the given table

    Example:
        >> columns([0])
        {item: int}
    """

    elems = obj.type.elems
    if isinstance(elems, tuple):    # Create a tuple/list instead of dict?
        elems = {f't{i}':e for i, e in enumerate(elems)}

    return ast.Dict_(elems)


def pql_cast(_state: State, obj: T.any, target_type: T.type):
    """Attempt to cast an object to a specified type

    The resulting object will be of type `target_type`, or a `TypeError`
    exception will be thrown.

    Parameters:
        obj: The object to cast
        target_type: The type to cast to

    """
    type_ = target_type
    if not isinstance(type_, Type):
        raise Signal.make(T.TypeError, type_, f"Cast expected a type, got {type_} instead.")

    if obj.type is type_:
        return obj

    return cast(obj, type_)


def pql_import_table(state: State, name: T.string, columns: T.list[T.string].as_nullable() = objects.null):
    """Import an existing table from the database, and fill in the types automatically.

    Parameters:
        name: The name of the table to import
        columns: If this argument is provided, only these columns will be imported.

    Example:
        >> import_table("my_sql_table", ["some_column", "another_column])
    """
    name_str = cast_to_python_string(state, name)
    assert isinstance(name_str, str)
    columns_whitelist = cast_to_python(state, columns)
    if columns_whitelist is not None:
        assert isinstance(columns_whitelist, list)
        columns_whitelist = set(columns_whitelist)

    # Get table type
    t = state.db.import_table_type(name_str, columns_whitelist)
    assert t <= T.table

    # Get table contents
    return objects.new_table(t, select_fields=bool(columns_whitelist))



def pql_connect(state: State, uri: T.string, load_all_tables: T.bool = ast.false, auto_create: T.bool = ast.false):
    """Connect to a new database, specified by the uri

    Parameters:
        uri: A string specifying which database to connect to (e.g. "sqlite:///test.db")
        load_all_tables: If true, loads all the tables in the database into the global namespace.
        auto_create: If true, creates the database if it doesn't already exist (Sqlite only)

    Example:
        >> connect("sqlite://:memory:")     // Connect to a database in memory
    """
    uri = cast_to_python_string(state, uri)
    load_all_tables = cast_to_python(state, load_all_tables)
    auto_create = cast_to_python(state, auto_create)
    state.connect(uri, auto_create=auto_create)
    if load_all_tables:
        state._py_api.load_all_tables()     # XXX
    return objects.null

def pql_help(_state: State, inst: T.any = objects.null):
    """Provides a brief summary for the given object
    """
    if inst is objects.null:
        text = (
            "To see the list of functions and objects available in the namespace, type '[b]names()[/b]'\n"
            "To see the next page of a table preview, type '.' and then enter\n"
            "\n"
            "To get help for a specific function, type '[b]help(an_object)[/b]'\n\n"
            "For example:\n"
            "    >> help(help)\n"
        )
        return pyvalue_inst(text, T.string).replace(type=T._rich)


    lines = []
    try:
        doc = autodoc(inst).print_text()    # TODO maybe html
        if doc:
            lines += [doc]
    except NotImplementedError:
        lines = [f"No help available for {inst.repr()}"]
    except runtype.DispatchError:
        lines += [f"<doc not available yet for object of type '{inst.type}>'"]
    except AutoDocError:
        lines += [f"<error generating documentation for object '{inst}'"]


    text = '\n'.join(lines) + '\n'
    return pyvalue_inst(text).replace(type=T._rich)

def _get_doc(v):
    s = ''
    if v.type <= T.function and v.docstring:
        s = v.docstring.splitlines()[0]
    return new_str(s).code

def pql_names(state: State, obj: T.any = objects.null):
    """List all names in the namespace of the given object.

    If no object is given, lists the names in the current namespace.
    """
    if obj is objects.null:
        all_vars = (state.ns.get_all_vars())
    else:
        all_vars = obj.all_attrs()

    assert all(isinstance(s, str) for s in all_vars)
    all_vars = list(all_vars.items())
    all_vars.sort()
    tuples = [sql.Tuple(T.list[T.string], [new_str(n).code,new_str(v.type).code, _get_doc(v)]) for n,v in all_vars]

    table_type = T.table(dict(name=T.string, type=T.string, doc=T.string))
    return objects.new_const_table(state, table_type, tuples)


def pql_tables(state: State):
    """Returns a table of all the persistent tables in the database.

    The resulting table has two columns: name, and type.
    """
    names = state.db.list_tables()
    values = [(name, state.db.import_table_type(name, None)) for name in names]
    tuples = [sql.Tuple(T.list[T.string], [new_str(n).code,new_str(t).code]) for n,t in values]

    table_type = T.table(dict(name=T.string, type=T.string))
    return objects.new_const_table(state, table_type, tuples)


def pql_env_vars(state: State):
    """Returns a table of all the environment variables.

    The resulting table has two columns: name, and value.
    """
    tuples = [sql.Tuple(T.list[T.string], [new_str(n).code,new_str(t).code]) for n,t in os.environ.items()]

    table_type = T.table({'name':T.string, 'value': T.string})
    return objects.new_const_table(state, table_type, tuples)



def create_internal_funcs(d):
    new_d = {}
    for names, f in d.items():
        if isinstance(names, str):
            names = (names,)
        for name in names:
            new_d[name] = create_internal_func(name, f)
    return new_d

breakpoint_funcs = create_internal_funcs({
    ('c', 'continue'): pql_brk_continue
})


def pql_exit(state, value: T.any.as_nullable() = None):
    """Exit the current interpreter instance.

    Can be used from running code, or the REPL.

    If the current interpreter is nested within another Preql interpreter (e.g. by using debug()),
    exit() will return to the parent interpreter.

    """
    raise ExitInterp(value)



def import_pandas(state, dfs):
    """Import pandas.DataFrame instances into SQL tables
    """
    import pandas as pd
    def normalize_item(i):
        if pd.isna(i):
            return None
        i = i.item() if hasattr(i, 'item') else i
        return i

    for name, df in dfs.items():
        if isinstance(df, pd.Series):
            cols = ['key', 'value']
            rows = [(dt.to_pydatetime() if isinstance(dt, datetime) else dt,v) for dt, v in df.items()]
        else:
            assert isinstance(df, pd.DataFrame)
            cols = list(df)
            rows = [[normalize_item(i) for i in rec]
                    for rec in df.to_records()]
            rows = [ row[1:] for row in rows ]    # drop index

        yield new_table_from_rows(state, name, cols, rows)

def pql_import_json(state: State, table_name: T.string, uri: T.string):
    """Imports a json file into a new table.

    Returns the newly created table.

    Parameters:
        table_name: The name of the table to create
        uri: A path or URI to the JSON file

    Note:
        This function requires the `pandas` Python package.
    """
    table_name = cast_to_python_string(state, table_name)
    uri = cast_to_python_string(state, uri)
    print(f"Importing JSON file: '{uri}'")

    import pandas
    df = pandas.read_json(uri)
    tbl ,= import_pandas(state, {table_name: df})
    return tbl



def pql_import_csv(state: State, table: T.table, filename: T.string, header: T.bool = ast.Const(T.bool, False)):
    """Import a csv file into an existing table

    Parameters:
        table: A table into which to add the rows.
        filename: A path to the csv file
        header: If true, skips the first line
    """
    # TODO better error handling, validation

    filename = cast_to_python_string(state, filename)
    header = cast_to_python(state, header)
    msg = f"Importing CSV file: '{filename}'"

    ROWS_PER_QUERY = 1024

    cons = TableConstructor.make(table.type)
    keys = []
    rows = []

    def insert_values():
        q = sql.InsertConsts2(table.type.options['name'], keys, rows)
        db_query(state, q)


    try:
        with open(filename, 'r', encoding='utf8') as f:
            line_count = len(list(f))
            f.seek(0)

            reader = csv.reader(f)
            for i, row in enumerate(rich.progress.track(reader, total=line_count, description=msg)):
                if i == 0:
                    matched = cons.match_params(state, row)
                    keys = [p.name for (p, _) in matched]

                if header and i == 0:
                    # Skip first line if header=True
                    continue

                values = ["'%s'" % (v.replace("'", "''")) for v in row]
                rows.append(values)

                if (i+1) % ROWS_PER_QUERY == 0:
                    insert_values()
                    rows = []

        if keys and rows:
            insert_values()

        return table
    except FileNotFoundError as e:
        raise Signal.make(T.FileError, None, str(e))


def _rest_func_endpoint(state, func):
    from starlette.responses import JSONResponse
    async def callback(request):
        params = [objects.pyvalue_inst(v) for k, v in request.path_params.items()]
        expr = ast.FuncCall(func, params)
        res = evaluate(state, expr)
        res = cast_to_python(state, res)
        return JSONResponse(res)
    return callback

def _rest_table_endpoint(state, table):
    from starlette.responses import JSONResponse
    async def callback(request):
        tbl = table
        params = dict(request.query_params)
        if params:
            conds = [ast.Compare('=', [ast.Name(k), objects.pyvalue_inst(v)])
                     for k, v in params.items()]
            expr = ast.Selection(tbl, conds)
            tbl = evaluate(state, expr)
        res = cast_to_python(state, tbl)
        return JSONResponse(res)
    return callback


def pql_serve_rest(state: State, endpoints: T.struct, port: T.int = pyvalue_inst(8080)):
    """Start a starlette server (HTTP) that exposes the current namespace as REST API

    Parameters:
        endpoints: A struct of type (string => function), mapping names to the functions.
        port: A port from which to serve the API

    Note:
        Requires the `starlette` package for Python. Run `pip install starlette`.

    Example:
        >> func index() = "Hello World!"
        >> serve_rest({index: index})
        INFO     Started server process [85728]
        INFO     Waiting for application startup.
        INFO     Application startup complete.
        INFO     Uvicorn running on http://127.0.0.1:8080 (Press CTRL+C to quit)
    """

    try:
        from starlette.applications import Starlette
        from starlette.responses import JSONResponse
        from starlette.routing import Route
    except ImportError:
        raise Signal.make(T.ImportError, None, "starlette not installed! Run 'pip install starlette'")

    try:
        import uvicorn
    except ImportError:
        raise Signal.make(T.ImportError, None, "uvicorn not installed! Run 'pip install uvicorn'")

    port_ = cast_to_python_int(state, port)

    async def root(_request):
        return JSONResponse(list(endpoints.attrs))

    routes = [
        Route("/", endpoint=root)
    ]

    for func_name, func in endpoints.attrs.items():
        path = "/" + func_name
        if func.type <= T.function:
            for p in func.params:
                path += "/{%s}" % p.name

            routes.append(Route(path, endpoint=_rest_func_endpoint(state, func)))
        elif func.type <= T.table:
            routes.append(Route(path, endpoint=_rest_table_endpoint(state, func)))
        else:
            raise Signal.make(T.TypeError, func, f"Expected a function or a table, got {func.type}")

    app = Starlette(debug=True, routes=routes)

    uvicorn.run(app, port=port_)
    return objects.null


internal_funcs = create_internal_funcs({
    'exit': pql_exit,
    'help': pql_help,
    'names': pql_names,
    'tables': pql_tables,
    'env_vars': pql_env_vars,
    'dir': pql_names,
    'connect': pql_connect,
    'import_table': pql_import_table,
    'count': pql_count,
    'temptable': pql_temptable,
    'table_concat': pql_table_concat,
    'table_intersect': pql_table_intersect,
    'table_union': pql_table_union,
    'table_subtract': pql_table_substract,
    'SQL': pql_SQL,
    'inspect_sql': pql_inspect_sql,
    'PY': pql_PY,
    'isa': pql_isa,
    'issubclass': pql_issubclass,
    'type': pql_type,
    'repr': pql_repr,
    'debug': pql_debug,
    '_breakpoint': pql_breakpoint,
    'get_db_type': pql_get_db_type,
    'cast': pql_cast,
    'columns': pql_columns,
    'import_csv': pql_import_csv,
    'import_json': pql_import_json,
    'serve_rest': pql_serve_rest,
    'force_eval': pql_force_eval,
    'fmt': pql_fmt,
})

joins = {
    'join': objects.InternalFunction('join', [], pql_join, objects.Param('tables')),
    'joinall': objects.InternalFunction('joinall', [], pql_joinall, objects.Param('tables')),
    'leftjoin': objects.InternalFunction('leftjoin', [], pql_leftjoin, objects.Param('tables')),
    'outerjoin': objects.InternalFunction('outerjoin', [], pql_outerjoin, objects.Param('tables')),
}



from typing import List, Any, Optional, Dict, Union
from dataclasses import field

from preql.utils import dataclass, TextReference, field_list

from . import pql_types as types
from .pql_types import T, Object
from .types_impl import pql_repr


# TODO We want Ast to typecheck, but sometimes types are still unknown (i.e. at parse time).
# * Use incremental type checks?
# * Use two tiers of Ast?

@dataclass
class Ast(Object):
    text_ref: Optional[TextReference] = field(init=False, default=None)

    def set_text_ref(self, text_ref):
        object.__setattr__(self, 'text_ref', text_ref)
        return self

class Expr(Ast):
    _args = ()

@dataclass
class Marker(Expr):
    pass

class Statement(Ast): pass


@dataclass
class Name(Expr):
    "Reference to an object (table, tabledef, column (in `where`), instance, etc.)"
    name: str

    def __repr__(self):
        return f'Name({self.name})'

@dataclass
class Parameter(Expr):
    "A typed object without a value"
    name: str
    type: types.Type

@dataclass
class ResolveParameters(Expr):
    obj: Object
    values: Dict[str, Object]


@dataclass
class ParameterizedSqlCode(Expr):
    type: Object
    string: Object
    # values: Dict[str, Object]
    # state: Any


@dataclass
class Attr(Expr):
    "Reference to an attribute (usually a column)"
    expr: Optional[Object] #Expr
    name: Union[str, Marker]

    _args = 'expr',

@dataclass
class Const(Expr):
    type: types.Type
    value: Any

    def repr(self):
        return pql_repr(self.type, self.value)

@dataclass
class Ellipsis(Expr):
    from_struct: Optional[Union[Expr, Marker]]
    exclude: List[Union[str, Marker]]

class BinOpExpr(Expr):
    _args = 'args',

class UnaryOpExpr(Expr):
    _args = 'expr',


@dataclass
class Compare(BinOpExpr):
    op: str
    args: List[Object]


@dataclass
class BinOp(BinOpExpr):
    op: str
    args: List[Object]

@dataclass
class Or(BinOpExpr):
    args: List[Object]

@dataclass
class And(BinOpExpr):
    args: List[Object]

@dataclass
class Not(UnaryOpExpr):
    expr: Object

@dataclass
class Neg(UnaryOpExpr):
    expr: Object

@dataclass
class Contains(BinOpExpr):
    op: str
    args: List[Object]

@dataclass
class DescOrder(Expr):
    value: Object

    _args = 'value',

@dataclass
class Range(Expr):
    start: Optional[Object]
    stop: Optional[Object]

    _args = 'start', 'stop'

@dataclass
class NamedField(Expr):
    name: Optional[str]
    value: Object #(Expr, types.PqlType)
    user_defined: bool = True

    _args = 'value',



class TableOperation(Expr): pass

@dataclass
class Selection(TableOperation):
    table: Object
    conds: List[Expr]

@dataclass
class Projection(TableOperation):
    table: Object
    fields: List[NamedField]
    groupby: bool = False
    agg_fields: List[NamedField] = field_list()

    def __post_init__(self):
        if self.groupby:
            assert self.fields or self.agg_fields
        else:
            assert self.fields and not self.agg_fields

@dataclass
class Order(TableOperation):
    table: Object
    fields: List[Expr]

@dataclass
class Update(TableOperation):
    table: Object
    fields: List[NamedField]

@dataclass
class Delete(TableOperation):
    table: Object
    conds: List[Expr]

@dataclass
class Slice(Expr):
    obj: Object
    range: Range

    _args = 'obj',

@dataclass
class New(Expr):
    type: str
    args: list   # Func args

@dataclass
class NewRows(Expr):
    type: str
    args: list   # Func args

@dataclass
class FuncCall(Expr):
    func: Any   # objects.Function ?
    args: list   # Func args



@dataclass
class One(Expr):
    expr: Object
    nullable: bool = False


@dataclass
class Type(Ast):
    type_obj: Object
    nullable: bool = False

class Definition:
    pass

@dataclass
class ColumnDef(Ast, Definition):
    name: str
    type: Type
    query: Optional[Expr] = None
    default: Optional[Expr] = None

@dataclass
class FuncDef(Statement, Definition):
    userfunc: Object   # XXX Why not use UserFunction?


@dataclass
class TableDef(Statement, Definition):
    name: str
    columns: List[Union[ColumnDef, Ellipsis]]
    methods: list


@dataclass
class TableDefFromExpr(Statement, Definition):
    name: str
    expr: Expr
    const: bool


@dataclass
class StructDef(Statement, Definition):
    name: str
    members: list

@dataclass
class SetValue(Statement):
    name: (Name, Attr)
    value: Expr

@dataclass
class InsertRows(Statement):
    name: (Name, Attr)
    value: Expr

@dataclass
class Print(Statement):
    value: List[Object]

@dataclass
class Assert(Statement):
    cond: Object

@dataclass
class Return(Statement):
    value: Object

@dataclass
class Throw(Statement):
    value: Object

@dataclass
class Import(Statement):
    module_path: str
    as_name: Optional[str] = None
    use_core: bool = True

@dataclass
class CodeBlock(Statement):
    statements: List[Ast]

@dataclass
class Try(Statement):
    try_: CodeBlock
    catch_expr: Expr
    catch_block: CodeBlock

@dataclass
class If(Statement):
    cond: Object
    then: Statement
    else_: Optional[Statement] = None

@dataclass
class For(Statement):
    var: str
    iterable: Object
    do: CodeBlock

@dataclass
class While(Statement):
    cond: Object
    do: Statement



# Collections

@dataclass
class List_(Expr):
    type: Object
    elems: list

@dataclass
class Table_Columns(Expr):
    type: Object
    cols: Dict[str, list]

@dataclass
class Dict_(Expr):
    elems: dict



def pyvalue(value):
    t = types.from_python(type(value))
    return Const(t, value)


false = Const(T.bool, False)
true = Const(T.bool, True)


from typing import Optional, List

from lark.exceptions import GrammarError

from preql.utils import dataclass, TextReference
from preql.context import context

from .base import Object


@dataclass
class Signal(Object, Exception):
    type: object    # Type
    text_refs: List[Optional[TextReference]]    # XXX must it be optional?
    message: Optional[str]

    @classmethod
    def make(cls, type, ast, message):
        ast_ref = getattr(ast, 'text_ref', None)
        try:
            refs = context.state.stacktrace+([ast_ref] if ast_ref else [])
        except AttributeError:
            refs = []
        return cls(type, refs, message)

    # def __str__(self):
    #     "Returns the exception with a traceback"
    #     texts = ['Exception traceback:\n']
    #     texts += [ref.get_pinpoint_text() if ref else '  ~~~ ???\n' for ref in self.text_refs]
    #     texts += [
    #         '[%s] %s\n' % (self.type, self.message)
    #     ]
    #     return ''.join(texts)

    def repr(self):
        return f'{self.type}("{self.message}")'


@dataclass
class pql_SyntaxError(GrammarError):
    text_ref: TextReference
    message: str

@dataclass
class ExitInterp(Exception):
    value: object

@dataclass
class ReturnSignal(Exception):
    value: object


@dataclass
class pql_AttributeError(Exception):
    message: str

class InsufficientAccessLevel(Exception):
    pass
class DatabaseQueryError(Exception):
    pass



import decimal
import json
from datetime import datetime
from typing import List, Optional, Dict

from preql.utils import dataclass, X, listgen, field_list, safezip
from preql.context import context

from . import pql_types
from .pql_types import T, Type, dp_type, dp_inst, Id
from .types_impl import join_names, flatten_type
from .exceptions import Signal

duck = 'duck'
sqlite = 'sqlite'
postgres = 'postgres'
bigquery = 'bigquery'
mysql = 'mysql'

class QueryBuilder:
    def __init__(self, target, is_root=True, start_count=0):
        self.target = target
        self.is_root = is_root

        self.counter = start_count

        self.table_name = []

    def unique_name(self):
        self.counter += 1
        return 't%d' % self.counter

    def replace(self, is_root):
        if is_root == self.is_root:
            return self # Optimize
        return QueryBuilder(self.target, is_root, self.counter)

    def push_table(self, t):
        self.table_name.append(t)
    def pop_table(self, t):
        t2 = self.table_name.pop()
        assert t2 == t

    def safe_name(self, base):
        "Return a name that is safe for use as variable. Must be consistent (pure func)"
        return _quote(self.target, base)

    def quote(self, id_):
        assert isinstance(id_, Id)
        return '.'.join(self.safe_name(n) for n in id_.parts)



class Sql:
    pass

@dataclass
class SqlTree(Sql):
    _is_select = False
    _needs_select = False

    def _compile(self, qb):
        raise NotImplementedError()

    def compile_wrap(self, qb):  # Move to Expr? Doesn't apply to statements
        return self.compile(qb).wrap(qb)

    def compile(self, qb):
        sql_code = self._compile(qb.replace(is_root=False))
        assert isinstance(sql_code, list), self
        assert all(isinstance(c, (str, Parameter)) for c in sql_code), self

        return CompiledSQL(self.type, sql_code, self, self._is_select, self._needs_select)

    def finalize_with_subqueries(self, qb, subqueries):
        if subqueries:
            subqs = [q.compile_wrap(qb).finalize(qb) for (name, q) in subqueries.items()]
            if qb.target in (postgres, mysql):
                sql_code = ['WITH RECURSIVE ']
            else:
                sql_code = ['WITH ']
            sql_code += join_comma([q, '\n    '] for q in subqs)
        else:
            sql_code = []
        sql_code += self.compile_wrap(qb).finalize(qb)
        return ''.join(sql_code)


@dataclass
class CompiledSQL(Sql):
    type: Type
    code: list
    source_tree: Optional[Sql]
    _is_select: bool   # Needed for embedding in SqlTree
    _needs_select: bool

    def finalize(self, qb):
        wrapped = self.wrap(qb)
        assert qb.is_root
        if wrapped.type <= T.primitive and not wrapped.code[0].lower().startswith('select '):
            code = ['SELECT '] + wrapped.code
        else:
            code = wrapped.code
        return ''.join(code)

    def wrap(self, qb):
        code = self.code

        if qb.is_root:
            if self._needs_select:
                code = ['SELECT * FROM '] + code
                return self.replace(code=code, _needs_select=False, _is_select=True)
        else:
            if self._is_select and not self._needs_select:
                # Bad recursion
                if qb.target in (postgres, mysql, duck):
                    code = ['('] + code + [') ', qb.unique_name()]  # postgres requires an alias
                else:
                    code = ['('] + code + [')']

                return self.replace(code=code, _is_select=False)

        return self

    def compile_wrap(self, qb):
        return self.wrap(qb)
    def compile(self, qb):
        return self
    def finalize_with_subqueries(self, qb, subqueries):
        return SqlTree.finalize_with_subqueries(self, qb, subqueries)

    def optimize(self):
        if not self.code:
            return self

        # unify strings for faster resolve_parameters and finalize
        new_code = []
        last = self.code[0]
        for c in self.code[1:]:
            if isinstance(c, str) and isinstance(last, str):
                last += c
            else:
                new_code.append(last)
                last = c
        new_code.append(last)
        return self.replace(code=new_code)


@dataclass
class RawSql(SqlTree):
    type: Type
    text: str

    def _compile(self, qb):
        return [self.text]

    @property
    def _is_select(self):
        return self.text.lstrip().lower().startswith('select')   # XXX Hacky! Is there a cleaner solution?

@dataclass
class Null(SqlTree):
    type = T.nulltype

    def _compile(self, qb):
        return ['null']

@dataclass
class Unknown(SqlTree):
    def _compile(self, qb):
        raise NotImplementedError("Unknown")

null = Null()
unknown = Unknown()

@dataclass
class Parameter(SqlTree):
    type: Type
    name: str

    def _compile(self, qb):
        return [self]


@dataclass
class Scalar(SqlTree):
    pass

@dataclass
class Atom(Scalar):
    pass

@dataclass
class Primitive(Atom):
    type: Type
    text: str

    def _compile(self, qb):
        return [self.text]


@dataclass
class Table(SqlTree):
    pass

@dataclass
class EmptyList(Table):
    type: Type

    _is_select = True

    def _compile(self, qb):
        return ['SELECT NULL AS ITEM LIMIT 0']

@dataclass
class TableName(Table):
    type: Type
    name: Id

    def _compile(self, qb):
        return [qb.quote(self.name)]

    _needs_select = True


class TableOperation(Table):
    _is_select = True


@dataclass
class FieldFunc(SqlTree):
    name: str
    field: Sql
    type = T.int

    def _compile(self, qb):
        return [f'{self.name}('] + self.field.compile_wrap(qb).code + [')']


@dataclass
class CountTable(Scalar):
    table: Sql
    type = T.int

    def _compile(self, qb):
        return [f'(SELECT COUNT(*) FROM '] + self.table.compile_wrap(qb).code + [')']

@dataclass
class JsonLength(Scalar):
    expr: Sql
    type = T.int

    def _compile(self, qb):
        code = self.expr.compile_wrap(qb).code
        if qb.target == sqlite:
            return [f'(length('] + code + [') - length(replace('] + code + [f', "{_ARRAY_SEP}", ""))) / length("{_ARRAY_SEP}") + 1']
        elif qb.target == postgres:
            return [f'array_length('] + code + [', 1)']
        else:
            return [f'json_length('] + code + [')']

@dataclass
class FuncCall(SqlTree):
    type: Type
    name: str
    fields: List[Sql]

    def _compile(self, qb):
        s = join_comma(f.compile_wrap(qb).code for f in self.fields)
        return [f'{self.name}('] + s + [')']

@dataclass
class Cast(SqlTree):
    type: Type
    value: Sql

    def _compile(self, qb):
        t = _compile_type(qb.target, self.type.as_nullable())     # XXX as-nullable here is a hack
        return [f'CAST('] + self.value.compile_wrap(qb).code + [f' AS {t})']


@dataclass
class Case(SqlTree):
    "SQL conditional"
    cond: Sql
    then: Sql
    else_: Optional[Sql]

    type = T.bool

    def _compile(self, qb):
        cond = self.cond.compile_wrap(qb).code
        then = self.then.compile_wrap(qb).code
        code = ["CASE WHEN "] + cond +[" THEN "] + then
        if self.else_:
            code += [ " ELSE " ] + self.else_.compile_wrap(qb).code
        return code + [" END "]

_ARRAY_SEP = '||'

@dataclass
class MakeArray(SqlTree):
    type: Type
    field: Sql

    def _compile(self, qb):
        field = self.field.compile_wrap(qb).code
        if qb.target == sqlite:
            return ['group_concat('] + field + [f', "{_ARRAY_SEP}")']
        elif qb.target == postgres:
            return ['array_agg('] + field + [')']
        elif qb.target == mysql:
            return ['json_arrayagg('] + field + [')']
        elif qb.target == bigquery:
            return ['array_agg('] + field + [')']

        assert False, qb.target


@dataclass
class Contains(Scalar):
    op: str
    exprs: List[Sql]
    type = T.bool

    def _compile(self, qb):
        assert self.op
        item, container = self.exprs
        c_item = item.compile_wrap(qb).code
        c_cont = container.compile_wrap(qb.replace(is_root=True)).code
        return c_item + [' ', self.op, ' '] + parens(c_cont)


def parens(x):
    return ['('] + x + [')']

@dataclass
class Compare(Scalar):
    op: str
    exprs: List[Sql]
    type = T.bool

    def __post_init__(self):
        assert self.op in ('=', '<=', '>=', '<', '>', '!=', 'IN'), self.op

    def _compile(self, qb):
        elems = [e.compile_wrap(qb).code for e in self.exprs]

        op = self.op

        if any(e.type.maybe_null() for e in self.exprs):
            # Null values are possible, so we'll use identity operators
            if qb.target == sqlite:
                op = {
                    '=': 'is',
                    '!=': 'is not'
                }.get(op, op)
            elif qb.target is mysql:
                if op == '!=':
                    # Special case,
                    return parens( ['not '] + join_sep(elems, f' <=> ') )

                op = {
                    '=': '<=>',
                }.get(op, op)
            else:
                op = {
                    '=': 'is not distinct from',
                    '!=': 'is distinct from'
                }.get(op, op)

        return parens( join_sep(elems, f' {op} ') )


@dataclass
class LogicalBinOp(Scalar):
    op: str
    exprs: List[Sql]

    def _compile(self, qb):
        x = join_sep([e.compile_wrap(qb).code for e in self.exprs], f' {self.op} ')
        return parens(x)

    type = T.bool

@dataclass
class LogicalNot(Scalar):
    expr: Sql

    def _compile(self, qb):
        x = ['NOT '] + self.expr.compile_wrap(qb).code
        return parens(x)

    type = T.bool


@dataclass
class BinOp(Scalar):
    op: str
    exprs: List[Sql]

    def _compile(self, qb):
        x = join_sep([e.compile_wrap(qb).code for e in self.exprs], f' {self.op} ')
        return parens(x)

    type = property(X.exprs[0].type)     # TODO ensure type correctness


@dataclass
class TableArith(TableOperation):
    op: str
    exprs: List[Table]

    def _compile(self, qb):
        tables = [t.compile_wrap(qb) for t in self.exprs]
        selects = [[f"SELECT * FROM "] + t.code for t in tables]

        code = join_sep(selects, f" {self.op} ")

        if qb.target == sqlite:
            # Limit -1 is due to a strange bug in SQLite (fixed in newer versions), where the limit is reset otherwise.
            code += [" LIMIT -1"]

        return code

    type = property(X.exprs[0].type)   # TODO ensure type correctness


@dataclass
class Neg(SqlTree):
    expr: Sql

    def _compile(self, qb):
        s = self.expr.compile_wrap(qb)
        return ["-"] + s.code

    type = property(X.expr.type)

@dataclass
class Desc(SqlTree):
    expr: Sql

    def _compile(self, qb):
        s = self.expr.compile_wrap(qb)
        return s.code + [" DESC"]

    type = property(X.expr.type)

_reserved = {'index', 'create', 'unique', 'table', 'select', 'where', 'group', 'by', 'over', 'user'}

@dataclass
class Name(SqlTree):
    type: Type
    name: str

    def __post_init__(self):
        assert self.name, self.type

    def _compile(self, qb):
        name = qb.safe_name(self.name)
        if qb.table_name:
            name = qb.table_name[-1] + '.' + name
        return [name]

@dataclass
class Attr(SqlTree):
    type: Type
    obj: Sql
    name: str



    # return base

@dataclass
class ColumnAlias(SqlTree):
    value: Sql
    alias: str

    @classmethod
    def make(cls, value, alias):
        return cls(value, alias)

    def _compile(self, qb):
        alias = qb.safe_name(self.alias)
        value = self.value.compile_wrap(qb).code
        assert alias and value, (alias, value)
        if value == [alias]:  # TODO disable when unoptimized?
            return value  # This is just for beauty, it's not necessary for function

        return value + [f' AS {alias}']

    type = property(X.value.type)


@dataclass
class Insert(SqlTree):
    table_name: Id
    columns: List[str]
    query: Sql
    type = T.nulltype

    def _compile(self, qb):
        columns = [qb.quote(Id(c)) for c in self.columns]
        return [f'INSERT INTO {qb.quote(self.table_name)}({", ".join(columns)}) '] + self.query.compile(qb).code

    def finalize_with_subqueries(self, qb, subqueries):
        if qb.target is mysql:
            columns = [qb.quote(Id(c)) for c in self.columns]
            sql_code = f'INSERT INTO {qb.quote(self.table_name)}({", ".join(columns)}) '
            sql_code += self.query.finalize_with_subqueries(qb, subqueries)
            return ''.join(sql_code)

        return super().finalize_with_subqueries(qb, subqueries)

@dataclass
class InsertConsts(SqlTree):
    table: str
    cols: List[str]
    tuples: list #List[List[Sql]]
    type = T.nulltype

    def _compile(self, qb):
        cols = self.cols
        tuples = self.tuples
        assert self.tuples, self

        values = join_comma(
            parens(join_comma([e.compile_wrap(qb).code for e in tpl]))
            for tpl in tuples
        )

        cols = [qb.quote(Id(c)) for c in cols]
        q = ['INSERT INTO', qb.safe_name(self.table),
             "(", ', '.join(cols), ")",
             "VALUES ",
        ]
        return [' '.join(q)] + values #+ [';']

@dataclass
class InsertConsts2(SqlTree):
    table: Id
    cols: List[str]
    tuples: list #List[List[Sql]]
    type = T.nulltype

    def _compile(self, qb):
        assert self.tuples, self

        values = join_comma(
            parens(join_comma([t] for t in tpl))
            for tpl in self.tuples
        )

        q = ['INSERT INTO', qb.quote(self.table),
             "(", ', '.join(self.cols), ")",
             "VALUES ",
        ]
        return [' '.join(q)] + values #+ ';'


@dataclass
class LastRowId(Atom):
    type = T.int

    def _compile(self, qb):
        if qb.target == sqlite:
            return ['last_insert_rowid()']   # Sqlite
        elif qb.target == mysql:
            return ['last_insert_id()']
        else:
            return ['lastval()']   # Postgres

@dataclass
class SelectValue(Atom, TableOperation):
    # XXX Just use a regular select?
    value: Sql

    def _compile(self, qb):
        value = self.value.compile_wrap(qb)
        return [f'SELECT '] + value.code + [' AS '] + value

    type = property(X.value.type)

# @dataclass
# class RowDict(SqlTree):
#     values: Dict[str, Sql]

#     def _compile(self, qb):
#         breakpoint()
#         return {v.compile_wrap(qb).code + [f' as {name}'] for name, v in self.values.items()}


@dataclass
class ValuesTuple(SqlTree):
    type: Type
    values: List[Sql]
    def _compile(self, qb):
        values = [v.compile_wrap(qb) for v in self.values]
        return join_comma(v.code for v in values)

@dataclass
class Values(Table):
    type: Type
    values: List[Sql]

    def _compile(self, qb):
        values = [v.compile_wrap(qb) for v in self.values]
        if not values:  # SQL doesn't support empty values
            nulls = ', '.join(['NULL' for _ in range(len(self.type.elems))])
            return ['SELECT ' + nulls + ' LIMIT 0']

        if qb.target == mysql:
            def row_func(x):
                return ['ROW('] + x + [')']
        else:
            row_func = parens

        return ['VALUES '] + join_comma(row_func(v.code) for v in values)

@dataclass
class Tuple(SqlTree):
    type: Type
    values: List[Sql]

    def _compile(self, qb):
        values = [v.compile_wrap(qb).code for v in self.values]
        return join_comma(values)

@dataclass
class ValuesTuples(Table):
    type: Type
    values: List[Tuple]

    def _compile(self, qb):
        if not self.values:  # SQL doesn't support empty values
            return ['SELECT '] + join_comma(['NULL']*len(self.type.elems)) +  ['LIMIT 0']
        values = [v.compile_wrap(qb) for v in self.values]
        return ['VALUES '] + join_comma(v.code for v in values)



@dataclass
class AllFields(SqlTree):
    type: Type

    def _compile(self, qb):
        return ['*']

@dataclass
class Update(SqlTree):
    table: TableName
    fields: Dict[Sql, Sql]
    conds: List[Sql]
    type = T.nulltype

    def _compile(self, qb):
        fields_sql = [k.compile_wrap(qb).code + [' = '] + v.compile_wrap(qb).code for k, v in self.fields.items()]
        fields_sql = join_comma(fields_sql)

        sql = ['UPDATE '] + self.table.compile_wrap(qb).code + [' SET '] + fields_sql

        if self.conds:
            sql += [' WHERE '] + join_sep([c.compile_wrap(qb).code for c in self.conds], ' AND ')

        return sql

@dataclass
class Delete(SqlTree):
    table: TableName
    conds: List[Sql]
    type = T.nulltype

    def _compile(self, qb):
        conds = join_sep([c.compile_wrap(qb).code for c in self.conds], ' AND ')
        return ['DELETE FROM '] + self.table.compile_wrap(qb).code + [' WHERE '] + conds

@dataclass
class Select(TableOperation):
    type: Type
    table: Sql # XXX Table won't work with RawSQL
    fields: List[Sql]
    conds: List[Sql] = field_list()
    group_by: List[Sql] = field_list()
    order: List[Sql] = field_list()

    # MySQL doesn't support arithmetic in offset/limit, and we don't need it anyway
    offset: Optional[int] = None
    limit: Optional[int] = None

    def __post_init__(self):
        assert self.fields, self

    def _is_conds_only(self):
        if self.group_by or self.order or self.offset or self.limit:
            return False

        if len(self.fields) == 1 and isinstance(self.fields[0], AllFields):
            return True

        return False

    def _compile(self, qb):
        # XXX very primitive optimization. Be smarter.
        #
        # Simplify
        #
        if isinstance(self.table, Select):
            s1 = self
            s2 = self.table
            if s2._is_conds_only():
                s = s1.replace(conds=list(s1.conds) + list(s2.conds), table=s2.table)
                return s._compile(qb)

            # elif s1._is_limit_only() and not (s2.offset or s2.limit):
            #     s = s2.replace(limit=s1.limit, offset=s1.offset)
            #     return s._compile(qb)

        #
        # Compile
        #
        fields_sql = [f.compile_wrap(qb) for f in self.fields]
        select_sql = join_comma(f.code for f in fields_sql)

        sql = ['SELECT '] + select_sql + [' FROM '] + self.table.compile_wrap(qb).code

        if self.conds:
            sql += [' WHERE '] + join_sep([c.compile_wrap(qb).code for c in self.conds], ' AND ')


        if self.group_by:
            sql += [' GROUP BY '] + join_comma(e.compile_wrap(qb).code for e in self.group_by)

        if self.order:
            sql += [' ORDER BY '] + join_comma(o.compile_wrap(qb).code for o in self.order)

        if self.limit is not None:
            sql += [' LIMIT ', str(self.limit)]
        elif self.offset is not None:
            if qb.target == sqlite:
                sql += [' LIMIT -1']  # Sqlite only (and only old versions of it)
            elif qb.target == mysql:
                # MySQL requires a specific limit, always!
                # See: https://stackoverflow.com/questions/255517/mysql-offset-infinite-rows
                sql += [' LIMIT 18446744073709551615']

        if self.offset is not None:
            sql += [' OFFSET ', str(self.offset)]

        return sql


@listgen
def join_sep(code_list, sep):
    code_list = list(code_list)
    if code_list:
        yield from code_list[0]
        for c in code_list[1:]:
            assert isinstance(c, list)
            yield sep
            yield from c

def join_comma(code_list):
    return join_sep(code_list, ", ")

@dataclass
class Subquery(SqlTree):
    table_name: str
    fields: Optional[List[Name]]
    query: Sql
    type = property(X.query.type)

    def _compile(self, qb):
        query = self.query.compile(qb).code
        if qb.target == bigquery:
            fields_str = []
        else:
            fields = [f.compile_wrap(qb.replace(is_root=False)).code for f in self.fields]
            fields_str = ["("] + join_comma(fields) + [")"] if fields else []
        return [f"{self.table_name}"] + fields_str + [" AS ("] + query + [")"]

def _enum_is_last(seq):
    last = len(seq) - 1
    for i, item in enumerate(seq):
        yield i == last, item


@dataclass
class Join(TableOperation):
    type: Type
    join_op: str
    tables: List[Table]
    conds: List[Sql]

    def _compile(self, qb):
        tables_sql = [t.compile_wrap(qb).code for t in self.tables]
        join_op = ' %s ' % self.join_op.upper()

        code = [f'SELECT * FROM '] + tables_sql[0]

        for is_last, t_sql in _enum_is_last(tables_sql[1:]):
            code += [join_op] + t_sql + [' ON ']

            if self.conds and is_last:
                code += join_sep([c.compile_wrap(qb).code for c in self.conds], ' AND ')
            else:
                code += ['1=1']   # Postgres requires ON clause

        return code



@dataclass
class BigQueryValues(SqlTree):
    type: Type
    values: List[Sql]

    def _compile(self, qb):
        cols = list(self.type.elems)

        rows = [
            (
             ['STRUCT('] + join_comma(v.compile(qb).code + [" as ", name]
                                  for name, v in safezip(cols, row.values)) + [")"]
            ) if isinstance(row, Tuple) else row.compile(qb).code
             for row in self.values
        ]

        return ["SELECT * FROM UNNEST(["] + join_comma(rows) + ["]) as item"]



@dataclass
class TableQueryValues(SqlTree):
    "A subquery which returns a table of given values"

    type: Type
    name: str
    rows: list

    def _compile(self, qb):
        if qb.target != 'bigquery':
            values_cls = Values
            fields = [Name(col_type, col_name) for col_name, col_type in self.type.elems.items()]
        else:
            values_cls = BigQueryValues
            fields = None

        subq = Subquery(self.name, fields, values_cls(self.type, self.rows))
        return subq._compile(qb)



@dataclass
class StringSlice(SqlTree):
    string: Sql
    start: Sql
    stop: Optional[Sql]

    type = T.string

    def _compile(self, qb):
        string = self.string.compile_wrap(qb).code
        start = self.start.compile_wrap(qb).code
        if self.stop:
            stop = self.stop.compile_wrap(qb).code
            length = parens(stop + ['-'] + start)
        else:
            length = None

        if qb.target == sqlite:
            f = 'substr'
            params = string + [', '] + start
            if length:
                params += [', '] + length
        elif qb.target in (postgres, mysql, bigquery):
            f = 'substring'
            params = string + [' from '] + start
            if length:
                params += [' for '] + length
        else:
            assert False

        return [f'{f}('] + params + [')']


@dp_type
def _repr(_t: T.number, x):
    return str(x)

@dp_type
def _repr(_t: T.bool, x):
    return ['false', 'true'][x]

@dp_type
def _repr(_t: T.decimal, x):
    return repr(float(x))  # TODO SQL decimal?

@dp_type
def _repr(_t: T.datetime, x):
    # TODO Better to pass the object instead of a string?
    return repr(str(x))

@dp_type
def _repr(_t: T.union[T.string, T.text], x):
    quoted_quote = r"\'" if context.state.db.target == bigquery else "''"
    return "'%s'" % str(x).replace("'", quoted_quote)

def _quote(target, name):
    assert isinstance(name, str), name
    if target is sqlite:
        return f'[{name}]'
    elif target is mysql or target is bigquery:
        return f'`{name}`'
    else:
        return f'"{name}"'



@dp_type
def _compile_type(target, type_: T.t_relation):
    # TODO might have a different type
    #return 'INTEGER'    # Foreign-key is integer
    return _compile_type(target, type_.elems['item'])

@dp_type
def _compile_type(target, type_: T.primitive):
    if target == bigquery:
        d = {
            'int': "INT64",
            'string': "STRING",
            'float': "FLOAT64",
            'bool': "BOOLEAN",
            'text': "STRING",
            'datetime': "TIMESTAMP",
        }
    elif target == mysql:
        d = {
            'int': "SIGNED",
            'string': "VARCHAR(4000)",
            'float': "FLOAT",
            'bool': "BOOLEAN",
            'text': "TEXT",
            'datetime': "TIMESTAMP",
        }
    else:
        d = {
            'int': "INTEGER",
            'string': "VARCHAR(4000)",
            'float': "FLOAT",
            'bool': "BOOLEAN",
            'text': "TEXT",
            'datetime': "TIMESTAMP",
        }
    s = d[type_.typename]
    if not type_.maybe_null():
        s += " NOT NULL"
    return s

@dp_type
def _compile_type(target, _type: T.nulltype):
    if target == bigquery:
        return 'INT64'
    return 'INTEGER'    # TODO is there a better value here? Maybe make it read-only somehow

@dp_type
def _compile_type(target, idtype: T.t_id):
    if target == bigquery:
        s = "STRING"
    else:
        s = "INTEGER"   # TODO non-int idtypes
    if not idtype.maybe_null():
        s += " NOT NULL"
    return s

@dp_type
def _compile_type(target, _type: T.json):
    return 'JSON'



def _from_datetime(s):
    if s is None:
        return None

    # Postgres
    if isinstance(s, datetime):
        return s

    # Sqlite
    if not isinstance(s, str):
        raise Signal.make(T.TypeError, None, f"datetime expected a string. Instead got: {s}")
    try:
        return datetime.fromisoformat(s)
    except ValueError as e:
        raise Signal.make(T.ValueError, None, str(e))


@dp_type
def _restructure_result(t, i):
    raise Signal.make(T.TypeError, None, f"Unexpected type used: {t}")

@dp_type
def _restructure_result(t: T.table, i):
    # return ({name: _restructure_result(state, col, i) for name, col in t.elem_dict.items()})
    return next(i)

@dp_type
def _restructure_result(t: T.struct, i):
    return ({name: _restructure_result(col, i) for name, col in t.elems.items()})

@dp_type
def _restructure_result(_t: T.union[T.primitive, T.nulltype], i):
    return next(i)


@dp_type
def _restructure_result(t: T.json_array[T.union[T.primitive, T.nulltype]], i):
    res = next(i)
    if not res:
        return res

    target = context.state.db.target
    if target == mysql:
        res = json.loads(res)
    elif target == sqlite:
        assert isinstance(res, str), res
        res = res.split(_ARRAY_SEP)

    # XXX hack! TODO Use a generic form to cast types
    if t.elem <= T.int:
        res = [int(x) for x in res]
    elif t.elem <= T.float:
        res = [float(x) for x in res]
    return res

@dp_type
def _restructure_result(_t: T.datetime, i):
    s = next(i)
    return _from_datetime(s)

# API

def compile_drop_table(state, table_name) -> Sql:
    target = state.db.target
    return RawSql(T.nulltype, f'DROP TABLE {_quote(target, table_name)}')


@dp_inst
def from_sql(res: T.primitive):
    try:
        row ,= res.value
        item ,= row
    except ValueError:
        raise Signal.make(T.TypeError, None, "Expected primitive. Got: '%s'" % res.value)
    # t = from_python(type(item))
    # if not (t <= res.type):
    #     raise Signal.make(T.TypeError, state, None, f"Incorrect type returned from SQL: '{t}' instead of '{res.type}'")
    return item


@dp_inst
def from_sql(res: T.datetime):
    # XXX doesn't belong here?
    row ,= res.value
    item ,= row
    s = item
    return _from_datetime(s)

def _from_sql_primitive(p):
    if isinstance(p, decimal.Decimal):
        # TODO Needs different handling when we expect a decimal
        return float(p)
    return p

@dp_inst
def from_sql(arr: T.list):
    fields = flatten_type(arr.type)
    if not all(len(e)==len(fields) for e in arr.value):
        raise Signal.make(T.TypeError, None, f"Expected 1 column. Got {len(arr.value[0])}")

    if arr.type.elem <= T.struct:
        return [{n: _from_sql_primitive(e) for (n, _t), e in safezip(fields, tpl)} for tpl in arr.value]
    else:
        return [_from_sql_primitive(e[0]) for e in arr.value]

@dp_inst
@listgen
def from_sql(arr: T.table):
    expected_length = len(flatten_type(arr.type))   # TODO optimize?
    for row in arr.value:
        if len(row) != expected_length:
            raise Signal.make(T.TypeError, None, f"Expected {expected_length} columns, but got {len(row)}")
        i = iter(row)
        yield {name: _restructure_result(col, i) for name, col in arr.type.elems.items()}


def compile_type_def(state, table_name, table) -> Sql:
    assert table <= T.table

    target = state.db.target

    posts = []
    pks = []
    columns = []

    pks = {join_names(pk) for pk in table.options['pk']}
    for name, c in flatten_type(table):
        if name in pks and c <= T.t_id:
            if target == postgres:
                type_ = "SERIAL" # Postgres
            elif target == mysql:
                type_ = "INTEGER NOT NULL AUTO_INCREMENT"
            elif target == bigquery:
                type_ = "STRING NOT NULL"
            else:
                type_ = "INTEGER"   # TODO non-int idtypes
        else:
            type_ = _compile_type(target if target != mysql else 'mysql_def', c)

        columns.append( f'{_quote(target, name)} {type_}' )

        if c <= T.t_relation:
            if state.db.target != 'bigquery':
                # TODO any column, using projection / get_attr
                if not table.options.get('temporary', False):
                    # In postgres, constraints on temporary tables may reference only temporary tables
                    rel = c.options['rel']
                    if rel['key']:          # Requires a unique constraint
                        _tbl_name ,= rel['table'].options['name'].parts   # TODO fix for multiple parts
                        s = f"FOREIGN KEY({name}) REFERENCES {_quote(target, _tbl_name)}({rel['column']})"
                        posts.append(s)

    if pks and target != bigquery:
        names = ", ".join(pks)
        posts.append(f"PRIMARY KEY ({names})")

    # Consistent among SQL databases
    command = "CREATE TEMPORARY TABLE" if table.options.get('temporary', False) else "CREATE TABLE IF NOT EXISTS"
    return RawSql(T.nulltype, f'{command} {_quote(target, table_name)} (' + ', '.join(columns + posts) + ')')




def deletes_by_ids(table, ids):
    for id_ in ids:
        compare = Compare('=', [Name(T.int, 'id'), Primitive(T.int, str(id_))])
        yield Delete(TableName(table.type, table.type.options['name']), [compare])

def updates_by_ids(table, proj, ids):
    sql_proj = {Name(value.type, name): value.code for name, value in proj.items()}
    for id_ in ids:
        compare = Compare('=', [Name(T.int, 'id'), Primitive(T.int, str(id_))])
        yield Update(TableName(table.type, table.type.options['name']), sql_proj, [compare])

def create_list(name, elems):
    # Assumes all elems have the same type!
    t = T.list[elems[0].type]
    subq = TableQueryValues(t, name, elems)
    table = TableName(t, Id(name))
    return table, subq, t



def create_table(t, name, rows):
    subq = TableQueryValues(t, name, rows)
    table = TableName(t, Id(name))
    return table, subq

def table_slice(table, start, stop):
    limit = stop - start if stop else None
    return Select(table.type, table.code, [AllFields(table.type)], offset=start, limit=limit)

def table_selection(table, conds):
    return Select(table.type, table.code, [AllFields(table.type)], conds=conds)

def table_order(table, fields):
    return Select(table.type, table.code, [AllFields(table.type)], order=fields)

def arith(target, res_type, op, args):
    arg_codes = list(args)
    if res_type == T.string:
        assert op == '+'
        op = '||'
        if target is mysql: # doesn't support a || b
            return FuncCall(res_type, 'concat', arg_codes)
    elif op == '/':
        if target != mysql:
            # In MySQL division returns a float. All others return int
            arg_codes[0] = Cast(T.float, arg_codes[0])
    elif op == '/~':
        if target == mysql:
            op = 'DIV'
        elif target == bigquery:
            # safe div?
            return FuncCall(res_type, 'div', arg_codes)
        else:
            op = '/'
    elif op == '%':
        if target is bigquery:
            return FuncCall(res_type, 'mod', arg_codes)
    elif op == '**':
        return FuncCall(T.float, 'power', arg_codes)

    return BinOp(op, arg_codes)

    
def make_value(x):
    if x is None:
        return null

    try:
        t = pql_types.from_python(type(x))
    except KeyError:
        raise ValueError(x)

    return Primitive(t, _repr(t, x))

def add_one(x):
    return BinOp('+', [x, make_value(1)])



from ast import literal_eval
from pathlib import Path

from lark import Lark, Transformer, v_args, UnexpectedInput, UnexpectedToken

from preql.utils import TextPos, TextRange, TextReference

from .exceptions import pql_SyntaxError
from . import pql_ast as ast
from . import pql_objects as objects
from .compiler import guess_field_name
from .pql_types import T


class Str(str):
    def __new__(cls, value, text_ref=None):
        obj = str.__new__(cls, value)
        obj.text_ref = text_ref
        return obj

    def set_text_ref(self, text_ref):
        # Required for parity with Ast
        self.text_ref = text_ref


def make_text_reference(text, source_file, meta, children=()):
    ref = TextRange(
        TextPos(
            meta.start_pos,
            meta.line,
            meta.column,
        ),
        TextPos(
            meta.end_pos or meta.start_pos,
            meta.end_line or meta.line,
            meta.end_column or meta.column,
        )
    )

    for c in children:
        if hasattr(c, 'text_ref'):
            assert c.text_ref.text is text
            assert c.text_ref.context is None
            c.text_ref.context = ref

    assert isinstance(source_file, (str, Path)), source_file
    return TextReference(text, str(source_file), ref)


def get_docstring(expr):
    if isinstance(expr, ast.CodeBlock):
        stmts = expr.statements
        if stmts and isinstance(stmts[0], ast.Const) and stmts[0].type is T.string:
            return stmts[0].value


# Taken from Lark (#TODO provide it in lark utils?)
def _fix_escaping(s):
    w = ''
    i = iter(s)
    for n in i:
        w += n
        if n == '\\':
            try:
                n2 = next(i)
            except StopIteration:
                raise ValueError("Literal ended unexpectedly (bad escaping): `%r`" % s)
            if n2 == '\\':
                w += '\\\\'
            elif n2 not in 'uxnftr':
                w += '\\'
            w += n2
    w = w.replace('\\"', '"').replace("'", "\\'")

    to_eval = "u'''%s'''" % w
    try:
        s = literal_eval(to_eval)
    except SyntaxError as e:
        raise ValueError(s, e)

    return s


def _wrap_result(res, f, meta, children):
    if isinstance(res, (Str, ast.Ast)):
        ref = make_text_reference(*f.__self__.code_ref, meta, children)
        res.set_text_ref(ref)
    return res

def _args_wrapper(f, _data, children, meta):
    "Create meta with 'code' from transformer"
    res = f(*children)
    return _wrap_result(res, f, meta, children)

def _args_wrapper_meta(f, _data, children, meta):
    ref = make_text_reference(*f.__self__.code_ref, meta, children)
    res = f(ref, *children)
    if isinstance(res, (Str, ast.Ast)):
        res.set_text_ref(ref)
    return res

def _args_wrapper_list(f, _data, children, meta):
    res = f(children)
    return _wrap_result(res, f, meta, children)

with_meta = v_args(wrapper=_args_wrapper_meta)
no_inline = v_args(wrapper=_args_wrapper_list)

def token_value(self, t):
    return Str(str(t))


@v_args(wrapper=_args_wrapper)
class TreeToAst(Transformer):
    def __init__(self, code_ref):
        super().__init__()
        self.code_ref = code_ref

    name = token_value

    def QUOTED_NAME(self, name):
        assert name[0] == name[-1] == '`'
        return name[1:-1]

    def STRING(self, s):
        return _fix_escaping(s[1:-1])
    def LONG_STRING(self, s):
        return _fix_escaping(s[3:-3])

    def string(self, s):
        return ast.Const(T.string, s)

    @with_meta
    def pql_dict(self, meta, items):
        d = {}
        for item in items:
            name = item.name or guess_field_name(item.value)
            if name in d:
                raise pql_SyntaxError(meta, f"Dict key appearing more than once: {name}")
            d[name] = item.value

        return ast.Dict_(d)

    def int(self, i):
        return ast.Const(T.int, int(i))

    def float(self, f):
        return ast.Const(T.float, float(f))

    def null(self):
        return ast.Const(T.nulltype, None)
    def false(self):
        return ast.Const(T.bool, False)
    def true(self):
        return ast.Const(T.bool, True)

    @no_inline
    def pql_list(self, items):
        return ast.List_(T.list[T.any], items)

    @no_inline
    def as_list(self, args):
        return args

    # types
    def typemod(self, *args):
        return [t.value for t in args]
    def type(self, name, mods):
        # TODO pk
        return ast.Type(name, '?' in (mods or ''))

    add_op = token_value
    mul_op = token_value
    comp_op = token_value

    def compare(self, a, op, b):
        return ast.Compare(op, [a,b])

    def _arith_expr(self, a, op, b):
        return ast.BinOp(op, [a,b])

    add_expr = _arith_expr
    term = _arith_expr
    power = lambda self, a,b: self._arith_expr(a, '**', b)

    def like(self, string, pattern):
        return ast.BinOp('like', [string, pattern])

    and_test = no_inline(ast.And)
    or_test = no_inline(ast.Or)

    not_test = ast.Not

    neg = ast.Neg
    var = ast.Name
    getattr = ast.Attr
    named_expr = ast.NamedField
    order = ast.Order
    update = ast.Update
    delete = ast.Delete
    desc = ast.DescOrder
    new = ast.New
    new_rows = ast.NewRows
    range = ast.Range

    selection = ast.Selection
    projection = ast.Projection
    slice = ast.Slice

    def projection_grouped(self, table, keys, values):
        return ast.Projection(table, keys, True, values)

    def projection_grouped_nokeys(self, table, values):
        return ast.Projection(table, [], True, values)

    def projection_grouped_novalues(self, table, keys):
        return ast.Projection(table, keys, True, [])

    # Statements / Declarations
    param = objects.Param
    param_variadic = objects.ParamVariadic

    @with_meta
    def func_def_short(self, meta, name, params, expr, doc):
        res = self.func_def(meta, name, params, expr)
        if doc:
            # Add docstring
            res = res.replace(userfunc=res.userfunc.replace(docstring=doc))
        return res

    @with_meta
    def func_def(self, meta, name, params, expr):
        collector = None
        for i, p in enumerate(params):
            if isinstance(p, objects.ParamVariadic):
                if i != len(params)-1:
                    msg = f"A variadic parameter may only appear at the end of the function ({p.name})"
                    raise pql_SyntaxError(meta, msg)

                collector = p
                params = params[:-1]


        docstring = get_docstring(expr)

        return ast.FuncDef(objects.UserFunction(name, params, expr, collector, docstring=docstring))

    @with_meta
    def func_call(self, meta, func, args):
        for i, a in enumerate(args):
            if isinstance(a, ast.Ellipsis):
                if i != len(args)-1:
                    raise pql_SyntaxError(meta, f"An inlined struct must appear at the end of the function call ({a})")


        return ast.FuncCall(func, args)

    def set_value(self, lval, rval):
        if not isinstance(lval, (ast.Name, ast.Attr)):
            raise pql_SyntaxError(lval.text_ref, f"{lval.type} is not a valid l-value")
        return ast.SetValue(lval, rval)

    insert_rows = ast.InsertRows
    struct_def = ast.StructDef
    table_def = ast.TableDef
    col_def = ast.ColumnDef
    print = no_inline(ast.Print)
    assert_ = ast.Assert
    return_stmt = ast.Return
    import_stmt = ast.Import
    throw = ast.Throw
    if_stmt = ast.If
    while_stmt = ast.While
    for_stmt = ast.For
    try_catch = ast.Try
    one = lambda self, nullable, expr: ast.One(expr, nullable is not None)

    def marker(self, _marker):
        return ast.Marker()

    def table_def_from_expr(self, const, name, table_expr):
        return ast.TableDefFromExpr(name, table_expr, const == 'const')

    codeblock = no_inline(ast.CodeBlock)

    def ellipsis(self, from_struct, *exclude):
        return ast.Ellipsis(from_struct, list(exclude))


    def __default__(self, data, children, meta):
        raise Exception("Unknown rule:", data)


class Postlexer:
    def process(self, stream):
        paren_level = []
        for token in stream:
            if not (paren_level and paren_level[-1] == 'LPAR' and token.type == '_NL'):
                assert token.end_pos is not None
                yield token

            if token.type in ('LPAR', 'LSQB', 'LBRACE'):
                paren_level.append(token.type)
            elif token.type in ('RPAR', 'RSQB', 'RBRACE'):
                p = paren_level.pop()
                assert p == 'L' + token.type[1:]

    # XXX Hack for ContextualLexer. Maybe there's a more elegant solution?
    @property
    def always_accept(self):
        return ('_NL',)


# Prevent expressions like (1and1) or (1ina)
# Changing these terminals in the grammar will prevent collision detection
# Waiting on interregular!
from lark.lexer import PatternRE
_operators = ['IN', 'NOT_IN', 'AND', 'OR']
def _edit_terminals(t):
    if t.name in _operators:
        t.pattern = PatternRE('%s(?!\w)' % t.pattern.value)

parser = Lark.open(
    'preql.lark',
    rel_to=__file__,
    parser='lalr',
    postlex=Postlexer(),
    start=['module', 'expr'],
    maybe_placeholders=True,
    propagate_positions=True,
    cache=True,
    edit_terminals=_edit_terminals,
)


def terminal_desc(name):
    if name == '_NL':
        return "<NEWLINE>"
    p = parser.get_terminal(name).pattern
    if p.type == 'str':
        return p.value
    return '<%s>' % name

def terminal_list_desc(term_list):
    return [terminal_desc(x) for x in term_list if x != 'MARKER']

def parse_stmts(s, source_file, wrap_syntax_error=True):
    try:
        tree = parser.parse(s+"\n", start="module")
    except UnexpectedInput as e:
        if not wrap_syntax_error:
            raise

        assert isinstance(source_file, (str, Path)), source_file

        pos =  TextPos(e.pos_in_stream, e.line, e.column)
        ref = TextReference(s, str(source_file), TextRange(pos, pos))
        if isinstance(e, UnexpectedToken):
            if e.token.type == '$END':
                msg = "Code ended unexpectedly"
                ref = TextReference(s, str(source_file), TextRange(pos, TextPos(len(s), -1 ,-1)))
            else:
                msg = "Unexpected token: %r" % e.token.value

            expected = e.accepts or e.expected
            if expected and len(expected) < 5:
                accepts = terminal_list_desc(expected)
                msg += '. Expected: %s' % ' or '.join(accepts)
        else:
            msg = "Unexpected character: %r" % s[e.pos_in_stream]

        raise pql_SyntaxError(ref, msg)

    return TreeToAst(code_ref=(s, source_file)).transform(tree)



from ipykernel.kernelbase import Kernel

import preql
from preql.core.autocomplete import autocomplete
from preql.core.pql_objects import null

from . import __version__

pql = preql.Preql()
pql.set_output_format('html')
display = pql._display

class PreqlKernel(Kernel):
    implementation = 'Preql'
    implementation_version = __version__
    language = 'go'
    language_version = preql.__version__
    language_info = {
        'name': 'Go',
        'mimetype': 'text/x-pql',
        'file_extension': '.pql',
        'pygments_lexer': 'go',
    }
    banner = "Preql"

    def do_execute(self, code, silent, store_history=True, user_expressions=None,
                   allow_stdin=False):
        if not silent:
            # r = requests.post("http://127.0.0.1:8080/html", code)
            # json = r.json()
            # res = pql(code)

            # Evaluate (Really just compile)
            with pql.interp.setup_context():
                try:
                    res = pql._run_code(code, '<jupyter>')

                    # Print
                    if res is null:
                        res = ''
                    elif res is not None:
                        res = res.repr()

                    json = {
                        'output': display.as_html() + str(res),
                        'success': True
                    }
                except preql.Signal as e:
                    display.print_exception(e)

                    json = {
                        'output': display.as_html(),
                        'success': False
                    }


            # if json['success']:
            #     stream_content = {'name': 'stdout', 'text': json['output']}
            # else:
            #     stream_content = {'name': 'stderr', 'text': json['output']}
            # self.send_response(self.iopub_socket, 'stream', stream_content)

            html = json['output']
            # self.send_info("Elapsed Time: {} !\n".format(elapsed_time))
            self.send_response(self.iopub_socket, 'display_data', {
                'data': {
                    "text/html": html,
                },
                "metadata": {
                    "image/png": {
                        "width": 640,
                        "height": 480,
                    },
                }
            })


        return {'status': 'ok',
                # The base class increments the execution count
                'execution_count': self.execution_count,
                'payload': [],
                'user_expressions': {},
               }


    def do_complete(self, code, cursor_pos):
        context, fragment = last_word(code[:cursor_pos])

        all_vars = autocomplete(pql.interp.state, context)
        matches = [f'{k}' for k in all_vars if k.startswith(fragment)]

        return {
            'status' : 'ok',
            'matches' : matches,
            'cursor_start' : cursor_pos - len(fragment),
            'cursor_end' : cursor_pos,
            'metadata' : {},
        }


def is_name(s):
    return s.isalnum() or s in ('_', '!')
def last_word(s):
    if not s:
        return '', ''
    i = len(s)
    while i and is_name(s[i-1]):
        i -= 1
    if i < len(s) and s[i] == '!' :
        i += 1  # hack to support ... !var and !in
    return s[:i], s[i:]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['preql', 'preql.core', 'preql.docstring', 'preql.jup_kernel']

package_data = \
{'': ['*'], 'preql': ['modules/*']}

install_requires = \
['dsnparse',
 'lark-parser>=0.11.1,<0.12.0',
 'prompt-toolkit',
 'pygments',
 'rich>=9.12.0,<10.0.0',
 'runtype>=0.1.7,<0.2.0']

extras_require = \
{'mysql': ['mysqlclient'], 'pgsql': ['psycopg2'], 'server': ['starlette']}

entry_points = \
{'console_scripts': ['preql = preql.__main__:main']}

setup_kwargs = {
    'name': 'preql-lang',
    'version': '0.2.1',
    'description': 'An interpreted relational query language that compiles to SQL',
    'long_description': '![alt text](logo_small.png "Logo")\n\nPreql is an interpreted, relational programming language, that specializes in database queries.\n\nIt is designed for use by data engineers, analysts and data scientists.\n\nPreql\'s main objective is to provide an alternative to SQL, in the form of a high-level programming language, with first-class functions, modules, strict typing, and Python integration.\n\n**How does it work?**\n\nPreql code is interpreted and gets compiled to SQL at runtime. This way, Preql gains the performance and abilities of SQL, but can also operate as a normal scripting language.\n\nCurrently supported dialects are:\n* Postgres\n* MySQL\n* Sqlite\n* BigQuery (WIP)\n* More... (planned)\n\nFor features that are database-specific, or aren\'t implemented in Preql, there is a `SQL()` function that provides a convenient escape hatch to write raw SQL code.\n\n**Main Features**\n\n* Modern syntax and semantics\n    - Interpreted, everything is an object\n    - Strong type system with gradual type validation and duck-typing\n* Compiles to SQL\n* Python and Pandas integration\n* Interactive shell (REPL) with auto-completion\n* Runs on Jupyter Notebook\n\n\n**Note: Preql is still work in progress, and isn\'t ready for production use, or any serious use quite yet.**\n\n[![tests](https://github.com/erezsh/preql/actions/workflows/tests.yml/badge.svg)](https://github.com/erezsh/Preql/actions/workflows/tests.yml)\n\n# Learn More\n\n- [**Read the documentation**](https://preql.readthedocs.io/en/latest/)\n\n- [Follow the tutorial](https://preql.readthedocs.io/en/latest/tutorial.html)\n\n- [Browse the examples](https://github.com/erezsh/Preql/tree/master/examples)\n\n- Check out the online demo: [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/erezsh/preql-binder/master?filepath=tutorial_exploring_a_database.ipynb)\n\n# Get started\n\nSimply install via pip:\n\n```sh\n    pip install -U preql-lang\n```\n\nThen just run the interpreter:\n\n```sh\n    preql\n```\n\nRequires Python 3.6+\n\n[Read more](https://preql.readthedocs.io/en/latest/getting-started.html)\n\n# Quick Example\n\n```javascript\n// Declare a new table\ntable Continent {\n    name: string\n    area: int       // km²\n    population: int\n}\n\n// Initialize the table, by inserting rows\nnew Continent("Africa", 30370000, 1287920000)\nnew Continent("Antarctica", 14000000, 4490)\nnew Continent("Asia", 44579000, 4545133000)\nnew Continent("Europe", 10180000, 742648000)\nnew Continent("North America", 24709000, 587615000)\nnew Continent("South America", 17840000, 428240000)\nnew Continent("Australia", 8600000, 41261000)\n\n// Print the continents, ordered by density\nprint Continent {\n    ...                         // Include existing fields\n    density: population / area  // Create new a field\n\n} order{^density}\n\n// Print the total land area\nprint "Total land area:", sum(Continent{area}), "km²"\n\n//  ========================= Output ==========================\n\n                              table  =7\n┏━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━┓\n┃ id ┃ name          ┃     area ┃ population ┃               density ┃\n┡━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━┩\n│  3 │ Asia          │ 44579000 │ 4545133000 │     101.9568182328002 │\n│  4 │ Europe        │ 10180000 │  742648000 │      72.9516699410609 │\n│  1 │ Africa        │ 30370000 │ 1287920000 │     42.40763911755021 │\n│  6 │ South America │ 17840000 │  428240000 │    24.004484304932735 │\n│  5 │ North America │ 24709000 │  587615000 │    23.781415678497712 │\n│  7 │ Australia     │  8600000 │   41261000 │     4.797790697674419 │\n│  2 │ Antarctica    │ 14000000 │       4490 │ 0.0003207142857142857 │\n└────┴───────────────┴──────────┴────────────┴───────────────────────┘\n\nTotal land area: 150278000 km²\n```\n\nIn the background, this table was generated by executing the following compiled SQL code (reformatted):\n\n```sql\n-- Continent {..., density: population / area} order{ ^density }\nWITH subq_1(id, name, area, population, density) AS (\n    SELECT id, name, area, population, (CAST(population AS float) / area) AS density\n    FROM Continent\n    ORDER BY density DESC)\nSELECT * FROM subq_1\n```\n\nSee more examples in [the examples folder](https://github.com/erezsh/Preql/tree/master/examples).\n\n# Interactive Environment\n\n![Screenshot.png](docs/screenshot.png)\n\n# License\n\nPreql uses an “Interface-Protection Clause” on top of the MIT license.\n\nSee: [LICENSE](LICENSE)\n\nIn simple words, it\'s free for personal use. Also, it can be used for any commercial or non-commercial purpose, as long as your product doesn\'t base its value on exposing the Preql language itself to your users. [Read more](https://preql.readthedocs.io/en/latest/faq.html#license)\n',
    'author': 'Erez Shin',
    'author_email': 'erezshin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/erezsh/Preql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from . import _base_imports

from .api import Preql, T
from .core.exceptions import Signal

# import importlib.metadata as importlib_metadata
# __version__ = importlib_metadata.version("prql")
__version__ = "0.2.1"



from pathlib import Path
from functools import wraps

from preql.utils import classify
from preql.context import context

from .exceptions import Signal, pql_SyntaxError, ReturnSignal
from .evaluate import State, execute, eval_func_call, import_module, evaluate, localize, cast_to_python
from .parser import parse_stmts
from . import pql_ast as ast
from . import pql_objects as objects
from .interp_common import pyvalue_inst, call_builtin_func
from .pql_types import T, Object
from .pql_functions import import_pandas
from .pql_functions import internal_funcs, joins


def initial_namespace():
    # TODO localinstance / metainstance
    ns = {k:v for k, v in T.items()}
    ns.update(internal_funcs)
    ns.update(joins)
    # TODO all exceptions
    name = '__builtins__'
    module = objects.Module(name, dict(ns))
    return [{name: module}]


def entrypoint(f):
    @wraps(f)
    def inner(interp, *args, **kwargs):
        with interp.setup_context():
            return f(interp, *args, **kwargs)
    return inner

class Interpreter:
    def __init__(self, sqlengine, display, use_core=True):
        self.state = State(self, sqlengine, display, initial_namespace())
        if use_core:
            mns = import_module(self.state, ast.Import('__builtins__', use_core=False)).namespace
            bns = self.state.get_var('__builtins__').namespace
            # safe-update
            for k, v in mns.items():
                if not k.startswith('__'):
                    assert k not in bns
                    bns[k] = v

    def setup_context(self):
        return context(state=self.state)

    def _execute_code(self, code, source_file, args=None):
        assert not args, "Not implemented yet: %s" % args
        try:
            stmts = parse_stmts(code, source_file)
        except pql_SyntaxError as e:
            raise Signal(T.SyntaxError, [e.text_ref], e.message)


        if stmts:
            if isinstance(stmts[0], ast.Const) and stmts[0].type == T.string:
                self.set_var('__doc__', stmts[0].value)

        last = None
        for stmt in stmts:
            try:
                last = execute(self.state, stmt)
            except ReturnSignal:
                raise Signal.make(T.CodeError, stmt, "'return' outside of function")

        return last

    def _include(self, fn, rel_to=None):
        if rel_to:
            fn = Path(rel_to).parent / fn
        with open(fn, encoding='utf8') as f:
            self._execute_code(f.read(), fn)

    def set_var(self, name, value):
        if not isinstance(value, Object):
            try:
                value = value._to_pql()
            except AttributeError:
                value = pyvalue_inst(value)

        self.state.set_var(name, value)

    def has_var(self, name):
        try:
            self.state.get_var(name)
        except Signal:
            return False
        return True



    #####################

    execute_code = entrypoint(_execute_code)
    include = entrypoint(_include)


    @entrypoint
    def evaluate_obj(self, obj):
        return evaluate(self.state, obj)

    @entrypoint
    def localize_obj(self, obj):
        return localize(self.state, obj)

    @entrypoint
    def call_func(self, fname, args):
        res = eval_func_call(self.state, self.state.get_var(fname), args)
        return evaluate(self.state, res)

    @entrypoint
    def cast_to_python(self, obj):
        return cast_to_python(self.state, obj)

    @entrypoint
    def call_builtin_func(self, name, args):
        return call_builtin_func(self.state, name, args)

    @entrypoint
    def import_pandas(self, dfs):
        return list(import_pandas(self.state, dfs))

    @entrypoint
    def list_tables(self):
        return self.state.db.list_tables()


    @entrypoint
    def load_all_tables(self):
        table_types = self.state.db.import_table_types()
        table_types_by_schema = classify(table_types, lambda x: x[0], lambda x: x[1:])

        for schema_name, table_types in table_types_by_schema.items():
            if schema_name:
                schema = objects.Module(schema_name, {})
                self.set_var(schema_name, schema)

            for table_name, table_type in table_types:
                db_name = table_type.options['name']
                inst = objects.new_table(table_type, db_name)

                if schema_name:
                    schema.namespace[table_name] = inst
                else:
                    if not self.has_var(table_name):
                        self.set_var(table_name, inst)


    def clone(self, use_core):
        state = self.state
        i = Interpreter(state.db, state.display, use_core=use_core)
        i.state.stacktrace = state.stacktrace   # XXX proper interface
        return i